from pydantic import BaseModel
from typing import Dict, List

from ksxt.models.common import GeneralResponse


class OrderBookData(BaseModel):
    side: str
    ob_num: int
    ob_price: float
    ob_qty: float


class OrderBookInfo(BaseModel):
    total_asks_qty: float
    total_bids_qty: float
    asks: List[OrderBookData]
    bids: List[OrderBookData]


class MultiSymbolOrderBookInfos(BaseModel):
    order_books: Dict[str, OrderBookInfo]


# Single symbol OrderBook에 대한 구체적인 Response 타입
class KsxtSingleOrderBookResponse(GeneralResponse[OrderBookInfo]):
    pass


# Multi symbol OrderBook에 대한 구체적인 Response 타입
class KsxtMultiOrderBookResponse(GeneralResponse[MultiSymbolOrderBookInfos]):
    pass



from dataclasses import dataclass
from datetime import datetime
import logging
from typing import Literal

from sqlalchemy import Boolean, Column, Date, Float, Integer, String
from ksxt.market.db import Base

from ksxt.market.us.stock import UsStockItem, UsStockMarket


# region typedef struct
# {
#     char    ncod[2+1];  /* National code        */
#     char    exid[3+1];  /* Exchange id          */
#     char    excd[3+1];  /* Exchange code        */
#     char    exnm[16+1]; /* Exchange name        */
#     char    symb[16+1]; /* Symbol               */
#     char    rsym[16+1]; /* realtime symbol      */
#     char    knam[64+1]; /* Korea name           */
#     char    enam[64+1]; /* English name         */
#     char    stis[1+1];  /* Security type        */
#                         /* 1:Index              */
#                         /* 2:Stock              */
#                         /* 3:ETP(ETF)           */
#                         /* 4:Warrant            */
#     char    curr[4+1];  /* currency             */
#     char    zdiv[1+1];  /* float position       */
#     char    ztyp[1+1];  /* data type            */
#     char    base[12+1]; /* base price           */
#     char    bnit[8+1];  /* Bid order size       */
#     char    anit[8+1];  /* Ask order size       */
#     char    mstm[4+1];  /* market start time(HHMM)  */
#     char    metm[4+1];  /* market end time(HHMM)    */
#     char    isdr[1+1];  /* DR 여부  :Y, N       */
#     char    drcd[2+1];  /* DR 국가코드          */
#     char    icod[4+1];  /* 업종분류코드         */
#     char    sjong[1+1]; /* 지수구성종목 존재 여부 */
#                         /*   0:구성종목없음      */
#                         /*   1:구성종목있음      */
#     char    ttyp[1+1];  /* Tick size Type        */
#     char    etyp[3+1]; /* 001: ETF 002: ETN 003: ETC 004: Others 005: VIX Underlying ETF 006: VIX Underlying ETN*/
#     char    ttyp_sb[3+1]; /* Tick size type 상세 (ttyp 9일 경우 사용) : 런던 제트라 유로넥스트  */
# }
# endregion typedef struct


@dataclass
class NasdaqItem(UsStockItem):
    CONTROL = [
        ("ncod", 2 + 1),
        ("exid", 3 + 1),
        ("excd", 3 + 1),
        ("exnm", 16 + 1),
        ("symb", 16 + 1),
        ("rsym", 16 + 1),
        ("knam", 64 + 1),
        ("enam", 64 + 1),
        ("stis", 1 + 1),
        ("curr", 4 + 1),
        ("zdiv", 1 + 1),
        ("ztyp", 1 + 1),
        ("base", 12 + 1),
        ("bnit", 8 + 1),
        ("anit", 8 + 1),
        ("mstm", 4 + 1),
        ("metm", 4 + 1),
        ("isdr", 1 + 1),
        ("drcd", 2 + 1),
        ("icod", 4 + 1),
        ("sjong", 1 + 1),
        ("ttyp", 1 + 1),
        ("etyp", 3 + 1),
        ("ttyp_sb", 3 + 1),
    ]

    # Notional code
    ncod: str
    # Exchange ID
    exid: str
    # Exchange Code
    excd: str
    # Exchange Name
    exnm: str
    # Symbol
    symb: str
    # Realtime Symbol
    rsym: str
    # Korean name
    knam: str
    # English name
    enam: str
    # Security Type
    stis: Literal["1", "2", "3", "4"]
    # currency
    curr: str
    # float position
    zdiv: str
    # data type
    ztyp: str
    # base price
    base: float
    # bid order size
    bnit: float
    # ask ordersize
    anit: float
    # market start time (HHMM)
    mstm: str
    # market end time (HHMM)
    metm: str
    # DR 여부 : Y/N
    isdr: bool
    # DR 국가코드
    drcd: str
    # 업종분류코드
    icod: str
    # 지수구성종목 존재 여부
    sjong: Literal["0", "1"]
    # Tick Size Type
    ttyp: str
    #
    etyp: Literal["001", "002", "003", "004", "005", "006"]
    # Tick size type 상세 (ttyp 8일 경우 사용) : 런던 제트라 유로넥스트
    ttyp_sb: str

    def __init__(self, data: str = None):
        if data:
            super().__init__(data)


class _NasdaqItem(Base):
    __tablename__ = "nasdaq"
    # Notional code
    ncod = Column(String)
    # Exchange ID
    exid = Column(String)
    # Exchange Code
    excd = Column(String)
    # Exchange Name
    exnm = Column(String)
    # Symbol
    symb = Column(String, primary_key=True)
    # Realtime Symbol
    rsym = Column(String)
    # Korean name
    knam = Column(String)
    # English name
    enam = Column(String)
    # Security Type
    stis = Column(String)
    # currency
    curr = Column(String)
    # float position
    zdiv = Column(String)
    # data type
    ztyp = Column(String)
    # base price
    base = Column(Float)
    # bid order size
    bnit = Column(Float)
    # ask ordersize
    anit = Column(Float)
    # market start time (HHMM)
    mstm = Column(String)
    # market end time (HHMM)
    metm = Column(String)
    # DR 여부 : Y/N
    isdr = Column(Boolean)
    # DR 국가코드
    drcd = Column(String)
    # 업종분류코드
    icod = Column(String)
    # 지수구성종목 존재 여부
    sjong = Column(String)
    # Tick Size Type
    ttyp = Column(String)
    #
    etyp = Column(String)
    # Tick size type 상세 (ttyp 8일 경우 사용) : 런던 제트라 유로넥스트
    ttyp_sb = Column(String)


class Nasdaq(UsStockMarket[NasdaqItem, _NasdaqItem]):
    def __init__(self, client):
        super().__init__(client, "nasdaq", "나스닥", "https://new.real.download.dws.co.kr/common/master/nasmst.cod.zip")



import ast


@staticmethod
def key_exists(dictionary, key):
    if hasattr(dictionary, "__getitem__") and not isinstance(dictionary, str):
        if isinstance(dictionary, list) and type(key) is not int:
            return False
        try:
            value = dictionary[key]
            return value is not None and value != ""
        except LookupError:
            return False
    return False


@staticmethod
def safe_value(dictionary, key, default_value=None):
    return dictionary[key] if key_exists(dictionary, key) else default_value


@staticmethod
def safe_string(dictionary, key, default_value=""):
    return str(dictionary[key]) if key_exists(dictionary, key) else default_value


@staticmethod
def safe_number(dictionary, key, default_value=0):
    value = safe_string(dictionary, key)
    if value == "":
        return default_value

    try:
        return float(value)
    except Exception:
        return default_value


@staticmethod
def safe_boolean(dictionary, key, default_value=False):
    value = safe_string(dictionary, key)
    if value == "":
        return default_value

    try:
        return bool(ast.literal_eval(value.capitalize()))
    except (ValueError, SyntaxError):
        return default_value



from dataclasses import dataclass
from datetime import datetime
import logging
from typing import Literal

from sqlalchemy import Boolean, Column, Date, Float, Integer, String
from ksxt.market.db import Base

from ksxt.market.krx.stock import KrxStockItem, KrxStockMarket


APNT_CLS_CODES = {
    "0": "미분류",
    "1": "건설기계",
    "2": "조선운송",
    "3": "철강소재",
    "4": "에너지화학",
    "5": "정보통신",
    "6": "금융",
    "7": "필수소비재",
    "8": "자유소비재",
}

# region typedef struct
# {
#     char    mksc_shrn_iscd[9];        /* 단축코드                                     */
#     char    stnd_iscd[12];             /* 표준코드                                     */
#     char    hts_kor_isnm[N];           /* 한글종목명                                   */
#     char    scrt_grp_cls_code[2];               /* 증권그룹구분코드                             */
#                                                 /* ST:주권 MF:증권투자회사 RT:부동산투자회사    */
#                                                 /* SC:선박투자회사 IF:사회간접자본투융자회사    */
#                                                 /* DR:주식예탁증서 EW:ELW EF:ETF                */
#                                                 /* SW:신주인수권증권 SR:신주인수권증서          */
#                                                 /* BC:수익증권 FE:해외ETF FS:외국주권           */
# 	char    avls_scal_cls_code[1];              /* 시가총액 규모 구분 코드 유가                 */
# 												/* (0:제외 1:대 2:중 3:소)                      */
#     char    bstp_larg_div_code[4];              /* 지수 업종 대분류 코드                        */
#     char    bstp_medm_div_code[4];              /* 지수 업종 중분류 코드                        */
#     char    bstp_smal_div_code[4];              /* 지수 업종 소분류 코드                        */
#     char    mnin_cls_code_yn[1];                /* 제조업 구분 코드 (Y/N)                       */
#     char    low_current_yn[1];               	/* 저유동성종목 여부 							*/
#     char    sprn_strr_nmix_issu_yn[1];          /* 지배 구조 지수 종목 여부 (Y/N)               */
#     char    kospi200_apnt_cls_code[1];          /* KOSPI200 섹터업종(20110401 변경됨) 			*/
#                                                 /* 0:미분류 1:건설기계 2:조선운송 3:철강소재 	*/
# 												/* 4:에너지화학 5:정보통신 6:금융 7:필수소비재 	*/
# 												/* 8: 자유소비재                                */
#     char    kospi100_issu_yn[1];                /* KOSPI100여부                                 */
#     char    kospi50_issu_yn[1];                 /* KOSPI50 종목 여부                            */
#     char    krx_issu_yn[1];                     /* KRX 종목 여부                                */
#     char    etp_prod_cls_code[1];            	/* ETP 상품구분코드								*/
# 												/* 0:해당없음 1:투자회사형 2:수익증권형			*/
# 												/* 3:ETN 4:손실제한ETN							*/
#     char    elw_pblc_yn[1];                     /* ELW 발행여부 (Y/N)                           */
#     char    krx100_issu_yn[1];                  /* KRX100 종목 여부 (Y/N)                       */
#     char    krx_car_yn[1];                      /* KRX 자동차 여부                              */
#     char    krx_smcn_yn[1];                     /* KRX 반도체 여부                              */
#     char    krx_bio_yn[1];                      /* KRX 바이오 여부                              */
#     char    krx_bank_yn[1];                     /* KRX 은행 여부                                */
#     char    etpr_undt_objt_co_yn[1];            /* 기업인수목적회사여부 						*/
#     char    krx_enrg_chms_yn[1];                /* KRX 에너지 화학 여부                         */
#     char    krx_stel_yn[1];                     /* KRX 철강 여부                                */
#     char    short_over_cls_code[1];             /* 단기과열종목구분코드 0:해당없음              */
#                                                 /* 1:지정예고 2:지정 3:지정연장(해제연기)       */
#     char    krx_medi_cmnc_yn[1];                /* KRX 미디어 통신 여부                         */
#     char    krx_cnst_yn[1];                     /* KRX 건설 여부                                */
#     char    krx_fnnc_svc_yn[1];                 /* 삭제됨(20151218)								*/
#     char    krx_scrt_yn [1];                    /* KRX 증권 구분                                */
#     char    krx_ship_yn [1];                    /* KRX 선박 구분                                */
#     char    krx_insu_yn[1];                     /* KRX섹터지수 보험여부                         */
#     char    krx_trnp_yn[1];                     /* KRX섹터지수 운송여부                         */
# 	char	sri_nmix_yn[1];                     /* SRI 지수여부 (Y,N)                           */
# 	char    stck_sdpr[9];                       /* 주식 기준가                                  */
#     char    frml_mrkt_deal_qty_unit[5];         /* 정규 시장 매매 수량 단위                     */
#     char    ovtm_mrkt_deal_qty_unit[5];         /* 시간외 시장 매매 수량 단위                   */
#     char    trht_yn[1];                         /* 거래정지 여부                                */
#     char    sltr_yn[1];                         /* 정리매매 여부                                */
#     char    mang_issu_yn[1];                    /* 관리 종목 여부                               */
#     char    mrkt_alrm_cls_code[2];              /* 시장 경고 구분 코드 (00:해당없음 01:투자주의 */
#                                                 /* 02:투자경고 03:투자위험                      */
# 	char    mrkt_alrm_risk_adnt_yn[1];          /* 시장 경고위험 예고 여부                      */
#     char    insn_pbnt_yn[1];                    /* 불성실 공시 여부                             */
#     char    byps_lstn_yn[1];                    /* 우회 상장 여부                               */
#     char    flng_cls_code[2];                   /* 락구분 코드 (00:해당사항없음 01:권리락       */
#                                                 /* 02:배당락 03:분배락 04:권배락 05:중간배당락  */
#                                                 /* 06:권리중간배당락 99:기타                    */
#                                                 /* S?W,SR,EW는 미해당(SPACE)                     */
#     char    fcam_mod_cls_code[2];               /* 액면가 변경 구분 코드 (00:해당없음           */
#                                                 /* 01:액면분할 02:액면병합 99:기타              */
#     char    icic_cls_code[2];                   /* 증자 구분 코드 (00:해당없음 01:유상증자      */
#                                                 /* 02:무상증자 03:유무상증자 99:기타)           */
#     char    marg_rate[3];                       /* 증거금 비율                                  */
#     char    crdt_able[1];                       /* 신용주문 가능 여부                           */
#     char    crdt_days[3];                       /* 신용기간                                     */
#     char    prdy_vol[12];                       /* 전일 거래량                                  */
#     char    stck_fcam[12];                      /* 주식 액면가                                  */
#     char    stck_lstn_date[8];                  /* 주식 상장 일자                               */
#     char    lstn_stcn[15];                      /* 상장 주수(천)                                */
#     char    cpfn[21];                           /* 자본금                                       */
#     char    stac_month[2];                      /* 결산 월                                      */
#     char    po_prc[7];                          /* 공모 가격                                    */
#     char    prst_cls_code[1];                   /* 우선주 구분 코드 (0:해당없음(보통주)         */
#                                                 /* 1:구형우선주 2:신형우선주                    */
#     char    ssts_hot_yn[1];                     /* 공매도과열종목여부  							*/
#     char    stange_runup_yn[1];                 /* 이상급등종목여부 							*/
#     char    krx300_issu_yn[1];                  /* KRX300 종목 여부 (Y/N)                       */
#     char    kospi_issu_yn[1];                   /* KOSPI여부                                    */
# 	char	sale_account[9];					/* 매출액                                       */
# 	char    bsop_prfi[9];						/* 영업이익                                     */
# 	char	op_prfi[9];							/* 경상이익                                     */
# 	char	thtr_ntin[5];						/* 당기순이익                                   */
# 	char	roe[9];								/* ROE(자기자본이익률)                          */
# 	char	base_date[8];						/* 기준년월                                     */
# 	char	prdy_avls_scal[9];					/* 전일기준 시가총액 (억)                       */

# 	char	grp_code[3];						/* 그룹사 코드                                  */
#     char    co_crdt_limt_over_yn[1];            /* 회사신용한도초과여부                         */
#     char    secu_lend_able_yn[1];               /* 담보대출가능여부                             */
#     char    stln_able_yn[1];                    /* 대주가능여부                                 */
# }   ST_KSP_CODE;
# endregion typedef struct


@dataclass
class KospiItem(KrxStockItem):
    CONTROL = [
        ("mksc_shrn_iscd", 9),
        ("stnd_iscd", 12),
        ("hts_kor_isnm", (9 + 12, -227)),
        ("_index", -1),
        ("_reverse", True),
        ("scrt_grp_cls_code", 2),
        ("avls_scal_cls_code", 1),
        ("bstp_larg_div_code", 4),
        ("bstp_medm_div_code", 4),
        ("bstp_smal_div_code", 4),
        ("mnin_cls_code_yn", 1),
        ("low_current_yn", 1),
        ("sprn_strr_nmix_issu_yn", 1),
        ("kospi200_apnt_cls_code", 1),
        ("kospi100_issu_yn", 1),
        ("kospi50_issu_yn", 1),
        ("krx_issu_yn", 1),
        ("etp_prod_cls_code", 1),
        ("elw_pblc_yn", 1),
        ("krx100_issu_yn", 1),
        ("krx_car_yn", 1),
        ("krx_smcn_yn", 1),
        ("krx_bio_yn", 1),
        ("krx_bank_yn", 1),
        ("etpr_undt_objt_co_yn", 1),
        ("krx_enrg_chms_yn", 1),
        ("krx_stel_yn", 1),
        ("short_over_cls_code", 1),
        ("krx_medi_cmnc_yn", 1),
        ("krx_cnst_yn", 1),
        ("_dummy", 1),
        ("krx_scrt_yn", 1),
        ("krx_ship_yn", 1),
        ("krx_insu_yn", 1),
        ("krx_trnp_yn", 1),
        ("sri_nmix_yn", 1),
        ("stck_sdpr", 9),
        ("frml_mrkt_deal_qty_unit", 5),
        ("ovtm_mrkt_deal_qty_unit", 5),
        ("trht_yn", 1),
        ("sltr_yn", 1),
        ("mang_issu_yn", 1),
        ("mrkt_alrm_cls_code", 2),
        ("mrkt_alrm_risk_adnt_yn", 1),
        ("insn_pbnt_yn", 1),
        ("byps_lstn_yn", 1),
        ("flng_cls_code", 2),
        ("fcam_mod_cls_code", 2),
        ("icic_cls_code", 2),
        ("marg_rate", 3),
        ("crdt_able", 1),
        ("crdt_days", 3),
        ("prdy_vol", 12),
        ("stck_fcam", 12),
        ("stck_lstn_date", 8),
        ("lstn_stcn", 15),
        ("cpfn", 21),
        ("stac_month", 2),
        ("po_prc", 7),
        ("prst_cls_code", 1),
        ("ssts_hot_yn", 1),
        ("stange_runup_yn", 1),
        ("krx300_issu_yn", 1),
        ("kospi_issu_yn", 1),
        ("sale_account", 9),
        ("bsop_prfi", 9),
        ("op_prfi", 9),
        ("thtr_ntin", 5),
        ("roe", 9),
        ("base_date", 8),
        ("prdy_avls_scal", 9),
        ("grp_code", 3),
        ("co_crdt_limt_over_yn", 1),
        ("secu_lend_able_yn", 1),
        ("stln_able_yn", 1),
    ]

    excd: str
    """거래소정보"""
    curr: str
    """화폐 정보"""

    mksc_shrn_iscd: str
    """단축코드"""
    stnd_iscd: str
    """표준코드"""
    hts_kor_isnm: str
    """한글종목명"""
    scrt_grp_cls_code: Literal["ST", "MF", "RT", "SC", "IF", "DR", "EW", "EF", "SW", "SR", "BC", "FE", "FS"]
    """증권그룹구분코드"""
    avls_scal_cls_code: Literal["0", "1", "2", "3"]
    """시가총액 규모 구분 코드 유가"""
    bstp_larg_div_code: str
    """지수 업종 대분류 코드"""
    bstp_medm_div_code: str
    """지수 업종 중분류 코드"""
    bstp_smal_div_code: str
    """지수 업종 소분류 코드"""
    mnin_cls_code_yn: bool
    """제조업 구분 코드"""
    low_current_yn: bool
    """저유동성종목 여부"""
    sprn_strr_nmix_issu_yn: bool
    """지배 구조 지수 종목 여부"""
    kospi200_apnt_cls_code: Literal["0", "1", "2", "3", "4", "5", "6", "7", "8"]
    """KOSPI200 섹터업종"""
    kospi100_issu_yn: bool
    """KOSPI100여부"""
    kospi50_issu_yn: bool
    """KOSPI50 종목 여부"""
    krx_issu_yn: bool
    """KRX 종목 여부"""
    etp_prod_cls_code: Literal["0", "1", "2", "3"]
    """ETP 상품구분코드"""
    elw_pblc_yn: bool
    """ELW 발행여부"""
    krx100_issu_yn: bool
    """KRX100 종목 여부"""
    krx_car_yn: bool
    """KRX 자동차 여부"""
    krx_smcn_yn: bool
    """KRX 반도체 여부"""
    krx_bio_yn: bool
    """KRX 바이오 여부"""
    krx_bank_yn: bool
    """KRX 은행 여부"""
    etpr_undt_objt_co_yn: bool
    """기업인수목적회사여부"""
    krx_enrg_chms_yn: bool
    """KRX 에너지 화학 여부"""
    krx_stel_yn: bool
    """KRX 철강 여부"""
    short_over_cls_code: Literal["0", "1", "2", "3"]
    """단기과열종목구분코드"""
    krx_medi_cmnc_yn: bool
    """KRX 미디어 통신 여부"""
    krx_cnst_yn: bool
    """KRX 건설 여부"""
    krx_scrt_yn: bool
    """KRX 증권 구분"""
    krx_ship_yn: bool
    """KRX 선박 구분"""
    krx_insu_yn: bool
    """KRX섹터지수 보험여부"""
    krx_trnp_yn: bool
    """KRX섹터지수 운송여부"""
    sri_nmix_yn: bool
    """SRI 지수여부"""
    stck_sdpr: int
    """주식 기준가"""
    frml_mrkt_deal_qty_unit: int
    """정규 시장 매매 수량 단위"""
    ovtm_mrkt_deal_qty_unit: int
    """시간외 시장 매매 수량"""
    trht_yn: bool
    """거래정지 여부"""
    sltr_yn: bool
    """정리매매 여부"""
    mang_issu_yn: bool
    """관리 종목 여부"""
    mrkt_alrm_cls_code: Literal["00", "01", "02", "03"]
    """시장 경고 구분 코드"""
    mrkt_alrm_risk_adnt_yn: bool
    """시장 경고위험 예고 여부"""
    insn_pbnt_yn: bool
    """불성실 공시 여부"""
    byps_lstn_yn: bool
    """우회 상장 여부"""
    flng_cls_code: Literal["00", "01", "02", "03", "04", "05", "06", "99"]
    """락구분 코드"""
    fcam_mod_cls_code: Literal["00", "01", "02", "99"]
    """액면가 변경 구분 코드"""
    icic_cls_code: Literal["00", "01", "02", "03", "99"]
    """증자 구분 코드"""
    marg_rate: int
    """증거금 비율"""
    crdt_able: bool
    """신용주문 가능 여부"""
    crdt_days: int
    """신용기간"""
    prdy_vol: int
    """전일 거래량"""
    stck_fcam: int
    """주식 액면가"""
    stck_lstn_date: datetime
    """주식 상장 일자"""
    lstn_stcn: int
    """상장 주수(천)"""
    cpfn: int
    """자본금"""
    stac_month: int
    """결산 월"""
    po_prc: int
    """공모 가격"""
    prst_cls_code: Literal["0", "1", "2"]
    """우선주 구분 코드"""
    ssts_hot_yn: bool
    """공매도과열종목여부"""
    stange_runup_yn: bool
    """이상급등종목여부"""
    krx300_issu_yn: bool
    """KRX300 종목 여부"""
    kospi_issu_yn: bool
    """KOSPI여부"""
    sale_account: int
    """매출액"""
    bsop_prfi: int
    """영업이익"""
    op_prfi: int
    """경상이익"""
    thtr_ntin: int
    """당기순이익"""
    roe: float
    """ROE(자기자본이익률)"""
    base_date: datetime
    """기준년월"""
    prdy_avls_scal: int
    """전일기준 시가총액 (억)"""
    grp_code: str
    """그룹사 코드"""
    co_crdt_limt_over_yn: bool
    """회사신용한도초과여부"""
    secu_lend_able_yn: bool
    """담보대출가능여부"""
    stln_able_yn: bool
    """대주가능여부"""

    @property
    def kospi200_apnt_cls_name(self) -> str:
        return APNT_CLS_CODES.get(self.kospi200_apnt_cls_code, None)

    def __init__(self, data: str = None):
        if data:
            super().__init__(data)


class _KospiItem(Base):
    """KOSPI 종목 정보"""

    __tablename__ = "kospi"
    excd = Column(String, default="kospi")
    curr = Column(String, default="KRW")

    mksc_shrn_iscd = Column(String, primary_key=True)
    stnd_iscd = Column(String)
    hts_kor_isnm = Column(String)
    scrt_grp_cls_code = Column(String)
    avls_scal_cls_code = Column(String)
    bstp_larg_div_code = Column(String)
    bstp_medm_div_code = Column(String)
    bstp_smal_div_code = Column(String)
    mnin_cls_code_yn = Column(Boolean)
    low_current_yn = Column(Boolean)
    sprn_strr_nmix_issu_yn = Column(Boolean)
    kospi200_apnt_cls_code = Column(String)
    kospi100_issu_yn = Column(Boolean)
    kospi50_issu_yn = Column(Boolean)
    krx_issu_yn = Column(Boolean)
    etp_prod_cls_code = Column(String)
    elw_pblc_yn = Column(Boolean)
    krx100_issu_yn = Column(Boolean)
    krx_car_yn = Column(Boolean)
    krx_smcn_yn = Column(Boolean)
    krx_bio_yn = Column(Boolean)
    krx_bank_yn = Column(Boolean)
    etpr_undt_objt_co_yn = Column(Boolean)
    krx_enrg_chms_yn = Column(Boolean)
    krx_stel_yn = Column(Boolean)
    short_over_cls_code = Column(String)
    krx_medi_cmnc_yn = Column(Boolean)
    krx_cnst_yn = Column(Boolean)
    krx_scrt_yn = Column(Boolean)
    krx_ship_yn = Column(Boolean)
    krx_insu_yn = Column(Boolean)
    krx_trnp_yn = Column(Boolean)
    sri_nmix_yn = Column(Boolean)
    stck_sdpr = Column(Integer)
    frml_mrkt_deal_qty_unit = Column(Integer)
    ovtm_mrkt_deal_qty_unit = Column(Integer)
    trht_yn = Column(Boolean)
    sltr_yn = Column(Boolean)
    mang_issu_yn = Column(Boolean)
    mrkt_alrm_cls_code = Column(String)
    mrkt_alrm_risk_adnt_yn = Column(Boolean)
    insn_pbnt_yn = Column(Boolean)
    byps_lstn_yn = Column(Boolean)
    flng_cls_code = Column(String)
    fcam_mod_cls_code = Column(String)
    icic_cls_code = Column(String)
    marg_rate = Column(Integer)
    crdt_able = Column(Boolean)
    crdt_days = Column(Integer)
    prdy_vol = Column(Integer)
    stck_fcam = Column(Integer)
    stck_lstn_date = Column(Date)
    lstn_stcn = Column(Integer)
    cpfn = Column(Integer)
    stac_month = Column(Integer)
    po_prc = Column(Integer)
    prst_cls_code = Column(String)
    ssts_hot_yn = Column(Boolean)
    stange_runup_yn = Column(Boolean)
    krx300_issu_yn = Column(Boolean)
    kospi_issu_yn = Column(Boolean)
    sale_account = Column(Integer)
    bsop_prfi = Column(Integer)
    op_prfi = Column(Integer)
    thtr_ntin = Column(Integer)
    roe = Column(Float)
    base_date = Column(Date)
    prdy_avls_scal = Column(Integer)
    grp_code = Column(String)
    co_crdt_limt_over_yn = Column(Boolean)
    secu_lend_able_yn = Column(Boolean)
    stln_able_yn = Column(Boolean)


class Kospi(KrxStockMarket[KospiItem, _KospiItem]):
    logger: logging.Logger

    def __init__(self, client):
        super().__init__(
            client, "kospi", "코스피", "https://new.real.download.dws.co.kr/common/master/kospi_code.mst.zip"
        )



import collections


@staticmethod
def sort_by(array, key, descending=False):
    return sorted(
        array, key=lambda k: getattr(k, key, None) if getattr(k, key, None) is not None else "", reverse=descending
    )



from datetime import datetime
from typing import Dict
from pydantic import BaseModel

from ksxt.models.common import GeneralResponse


class TickerInfo(BaseModel):
    symbol: str
    price: float
    ts: datetime


class MultiSymbolTickerInfo(BaseModel):
    tickers: Dict[str, TickerInfo]


# 단일 Ticker에 대한 Response 타입
class KsxtTickerResponse(GeneralResponse[TickerInfo]):
    pass


# 복수 Ticker에 대한 Response 타입
class KsxtTickersResponse(GeneralResponse[MultiSymbolTickerInfo]):
    pass



from abc import abstractmethod
from datetime import datetime, timedelta
import io
import logging
from threading import Lock
import zipfile
import requests
from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar, get_args
from sqlalchemy.orm import Session
from sqlalchemy import Column, and_

from ksxt.market.db import Base, Market
from ksxt.market.logging import KsxtLogger

if TYPE_CHECKING:
    from ksxt.market.manager import MarketManager


class MarketItemBase:
    """시장 종목"""

    def __init__(self, data: str, delimiter: Optional[str] = None):
        items = self.__class__.__dict__["CONTROL"].copy()
        data_len = len(data)
        index = 0
        reverse = False

        if delimiter is not None:
            splitted = data.split(delimiter)

        while items:
            key, value = items[0]
            del items[0]

            if key == "_reverse":
                items.reverse()
                reverse = not reverse
                continue
            elif key == "_index":
                if value < 0:
                    value = data_len - (abs(value) - 1)
                index = value
                continue
            elif key[:2] == "__":
                continue
            elif isinstance(value, tuple):
                value = list(value)

                for i, v in enumerate(value):
                    if v < 0:
                        value[i] = data_len - abs(v)

                v = data[value[0] : value[1]]
            elif reverse:
                v = data[index - value : index]
                index -= value
            else:
                if delimiter is None:
                    v = data[index : index + value]
                    index += value
                else:
                    v = splitted[index]
                    index += 1

            if key == "_dummy":
                continue

            type = self.__annotations__[key]
            setattr(self, key, self.__convert(type, v.strip()))

    def __convert(self, type, value):
        try:
            if type != str:
                if type == bool:
                    return value == "Y" if value else None
                elif type == datetime:
                    return datetime.strptime(value, "%Y%m%d")
                elif type.__name__ == "Literal":
                    return value

                return type(value)
        except:
            return None
        return value


t_item = TypeVar("t_item", bound=MarketItemBase)
t_dbitem = TypeVar("t_dbitem", bound=Base)


class MarketBase(Generic[t_item, t_dbitem], KsxtLogger):
    manager: "MarketManager"
    code: str
    name: str
    mst: str
    encoding: str = "cp949"

    _lock: Lock

    def __init__(self, manager: "MarketManager", code: str, name: str, mst: str):
        self.manager = manager
        self.code = code
        self.name = name
        self.mst = mst
        self._lock = Lock()

    def sync(self, cache: bool = True):
        """종목 정보를 동기화합니다."""
        with self._lock:
            if cache:
                sync_at = self.manager.sync_at(self.code)

                if sync_at and datetime.now() - sync_at < timedelta(days=1):
                    self.manager.logger.info(f"MARKET: up to date {self.code}")
                    return

            self._sync()

    def _sync(self):
        self.logger.info(f"MARKET: sync {self.code}, download {self.mst}")
        with zipfile.ZipFile(io.BytesIO(requests.get(self.mst).content)) as archive:
            data = archive.read(archive.filelist[0])
        self._update(data.decode(self.encoding))

    def _update(self, data: str):
        lines = data.splitlines()
        self.logger.info(f"MARKET: parsing {self.code} data... {len(lines)} lines")
        type, db_type = get_args(self.__orig_bases__[0])  # type: ignore

        with self.manager.session as ss:
            for line in lines:
                if not line:
                    continue
                item = db_type()
                state = item._sa_instance_state
                item.__dict__ = type(line).__dict__
                item._sa_instance_state = state
                ss.merge(item)

            self.manager._update_sync_at(self.code)
            ss.commit()

            if self.logger.level == logging.DEBUG:
                self.logger.info(f"MARKET: up to date {self.code}, {ss.query(db_type).count()} items")

    def _sf_acc(self):
        with self._lock:
            pass

    def _market(self) -> Market:
        return Market(code=self.code, name=self.name)

    @property
    def _session(self) -> Session:
        return self.manager.session

    def _search(self, column: Column, name: str, limit: int = 50) -> Iterable:
        self._sf_acc()
        type, db_type = get_args(self.__orig_bases__[0])  # type: ignore

        with self._session as sess:
            keywords = name.split()
            for item in (
                sess.query(db_type)
                .filter(and_(*[column.like(f"%{keyword}%") for keyword in keywords]))
                .limit(limit)
                .all()
            ):
                data = type()
                del item.__dict__["_sa_instance_state"]
                data.__dict__ = item.__dict__
                yield data

    def _get(self, column: Column, data: str) -> t_item:
        self._sf_acc()
        type, db_type = get_args(self.__orig_bases__[0])  # type: ignore

        with self._session as sess:
            item = sess.query(db_type).filter(column == data).first()

            if item:
                data = type()
                del item.__dict__["_sa_instance_state"]
                data.__dict__ = item.__dict__
                return data  # type: ignore

            return None

    def items(self, offset: int = 0, limit: int = 100) -> Iterable:
        """종목 정보를 반환합니다."""
        self._sf_acc()
        type, db_type = get_args(self.__orig_bases__[0])  # type: ignore

        with self._session as sess:
            for item in sess.query(db_type).offset(offset).limit(limit).all():
                data = type()
                del item.__dict__["_sa_instance_state"]
                data.__dict__ = item.__dict__
                yield data

    def all(self):
        self._sf_acc()
        return self.items(limit=999999)

    @abstractmethod
    def search(self, name: str, limit: int = 50) -> Iterable:
        """종목 이름으로 검색합니다."""
        raise NotImplementedError()

    def search_one(self, name: str) -> t_item:
        for i in self.search(name, 1):
            return i

        return None



from typing import Union
from ksxt.market.krx.kospi import Kospi, KospiItem
from ksxt.market.krx.kosdaq import Kosdaq, KosdaqItem
from ksxt.market.us.nasdaq import Nasdaq, NasdaqItem
from ksxt.market.us.nyse import Nyse, NyseItem
from ksxt.market.us.amex import Amex, AmexItem

MARKETS = {
    "kospi": Kospi,
    "kosdaq": Kosdaq,
    "nyse": Nyse,
    "nasdaq": Nasdaq,
    "amex": Amex,
}

MARKET_TYPE = Union[Kospi, Kosdaq, Nyse, Nasdaq, Amex]
MARKET_ITEM_TYPE = Union[KospiItem, KosdaqItem, NyseItem, NasdaqItem, AmexItem]



import importlib


def generate_class_code(exchange_name: str):
    try:
        # Dynamically import the module based on the exchange name
        module = importlib.import_module(f"{exchange_name}")
        exchange_methods = module.entries[exchange_name]
    except ImportError:
        raise ValueError(f"Module for exchange '{exchange_name}' could not be found.")
    except AttributeError:
        raise ValueError(f"Entries for exchange '{exchange_name}' not found in the module.")

    class_code = "class ImplicitAPI:\n"

    # Calculate maximum length for method names for alignment
    max_length = 0
    for category, methods in exchange_methods.items():
        for access, http_methods in methods.items():
            if isinstance(http_methods, dict):
                for method, paths in http_methods.items():
                    max_length = max(max_length, *(len(f"{access}_{method}_{path}") for path in paths))
            elif isinstance(http_methods, list):
                max_length = max(max_length, *(len(f"{access}_{path}") for path in http_methods))

    for category, methods in exchange_methods.items():
        class_code += f"\n    # {category} API methods\n"
        for access, http_methods in methods.items():
            class_code += f"\n    # {access} methods\n"
            if isinstance(http_methods, dict):
                for method, paths in http_methods.items():
                    class_code += f"\n    # {method} requests\n"
                    for path in paths:
                        entry_name = f"{access}_{method}_{path}"
                        # Right-align entry names for better readability
                        class_code += f"    {entry_name.ljust(max_length)} = Entry('{category}', '{path}', {{}})\n"
            elif isinstance(http_methods, list):
                for path in http_methods:
                    entry_name = f"{access}_{path}"
                    class_code += f"    {entry_name.ljust(max_length)} = Entry('{category}', '{path}', {{}})\n"

    return class_code


if __name__ == "__main__":
    exchange_names = ["bithumb", "koreainvest", "upbit"]  # List of exchanges to generate files for

    for exchange_name in exchange_names:
        class_code = generate_class_code(exchange_name)
        filename = f"./src/ksxt/api/{exchange_name}.py"
        with open(filename, "w") as file:
            file.write("from ksxt.api import Entry\n")
            file.write("\n")
            file.write(class_code)



entries = {
    "upbit": {
        "stock": {
            "public": {
                "get": [
                    "fetch_markets",
                    "fetch_security_ohlcv_minute",
                    "fetch_security_ohlcv_day",
                    "fetch_security_ohlcv_week",
                    "fetch_security_ohlcv_month",
                    "fetch_ticker_price",
                    "fetch_tickers_price",
                    "fetch_orderbook",
                    "fetch_orderbooks",
                ]
            },
            "private": {
                "get": [
                    "fetch_balance",
                    "fetch_cash",
                    "fetch_security_info",
                    "fetch_trade_fee",
                    "fetch_closed_order_detail",
                    "fetch_opened_order_detail",
                    "fetch_opened_order",
                    "fetch_closed_order",
                    "fetch_withdrawal_history",
                    "fetch_deposit_history",
                ],
                "post": ["send_order_entry", "send_order_entry_market", "send_order_exit", "send_order_exit_market"],
                "delete": ["send_cancel_order"],
            },
        },
        "websocket": {
            "public": ["subscribe_ticker", "subscribe_trade", "subscribe_orderbook", "subscribe_order"],
            "private": ["subscribe_ticker", "subscribe_trade"],
        },
    }
}



from sqlalchemy import Column, String, DateTime
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Market(Base):
    __tablename__ = "market"

    # Market Code
    code = Column(String, primary_key=True)

    # Market Name
    name = Column(String, nullable=False)

    # Sync time
    sync_at = Column(DateTime, nullable=True)



from datetime import datetime, timedelta
import json
import os
import time
from typing import Any, Dict, Literal, Optional

import pytz

from ksxt.api.koreainvest import ImplicitAPI
from ksxt.async_.base.async_exchange import AsyncExchange
from ksxt.market.manager import MarketManager
from ksxt.parser.koreainvest import KoreaInvestParser

import ksxt.models


class KoreaInvest(AsyncExchange, ImplicitAPI):
    def __init__(self, config: Dict = None) -> None:
        super().__init__(config, "koreainvest.toml")
        self.parser = KoreaInvestParser()
        self.timezone = pytz.timezone("Asia/Seoul")

    def is_activate(self, path, security_type) -> bool:
        mode = "dev" if self.is_dev == True else "app"

        tr_id = self.apis[self.type][security_type][path][mode]["tr_id"]

        if security_type != "token" and not bool(tr_id):
            return False

        return super().is_activate(path=path, security_type=security_type)

    def sign(
        self,
        path,
        security_type,
        method_type,
        api_type: Any = "public",
        headers: Optional[Any] = None,
        body: Optional[Any] = None,
        params: Optional[Any] = None,
        config={},
    ):
        mode = "dev" if self.is_dev == True else "app"

        host_url = self.apis[self.type][mode]["hostname"]
        destination = self.apis[self.type][security_type][path]["url"]
        version = self.apis[self.type]["version"]
        params["version"] = version
        destination = self.implode_params(destination, params)

        url = f"{host_url}/{destination}"

        tr_id = self.apis[self.type][security_type][path][mode]["tr_id"]
        authorization_token = f"Bearer {self.token}"

        if api_type == "private":
            if headers is None:
                headers = {}
                headers.update(
                    {
                        "content-type": "application/json",
                        "authorization": authorization_token,
                        "appkey": self.open_key,
                        "appsecret": self.secret_key,
                        "tr_id": tr_id,
                        "custtype": "P",
                    }
                )

        if method_type.upper() == "POST":
            body = json.dumps(params)
            params = {}

        return {"url": url, "method": method_type, "headers": headers, "body": body, "params": params}

    async def create_token(self) -> ksxt.models.KsxtTokenResponse:
        params = {"grant_type": "client_credentials", "appkey": self.open_key, "appsecret": self.secret_key}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_post_generate_token(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTokenResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_token(response=response)

        self.save_token(self.open_key, parsed_info.access_token, expired=parsed_info.expired_datetime)

        return ksxt.models.KsxtTokenResponse(header=common_header, response=common_response, info=parsed_info)

    def get_common_response(self, response):
        if "error_code" in response:
            return self.create_common_response(
                success="1",
                msg_code=self.safe_string(response, "error_code"),
                msg=self.safe_string(response, "error_description"),
                info=response,
            )

        if "rt_cd" in response and response["rt_cd"] != "0":
            return self.create_common_response(
                success="1",
                msg_code=self.safe_string(response, "msg_cd"),
                msg=self.safe_string(response, "msg1"),
                info=response,
            )

        if "response" in response and response["response"]["success"] != "0":
            return self.create_common_response(
                success="1",
                msg_code=self.safe_string(response["response"], "code"),
                msg=self.safe_string(response["response"], "message"),
                info=response,
            )

        return self.create_common_response(
            success="0",
            msg_code=self.safe_string(response, "msg_cd"),
            msg=self.safe_string(response, "msg1"),
            info=response,
        )

    @AsyncExchange.check_token
    async def fetch_balance(
        self, acc_num: str, base_market: str = "KRW", excluded_symbols: Optional[list[str]] = None
    ) -> ksxt.models.KsxtBalanceResponse:
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "AFHR_FLPR_YN": "N",
                "OFL_YN": "",
                "INQR_DVSN": "01",
                "UNPR_DVSN": "01",
                "FUND_STTL_ICLD_YN": "N",
                "FNCG_AMT_AUTO_RDPT_YN": "N",
                "PRCS_DVSN": "01",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_balance(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_balance(
            response=response, base_market=base_market, excluded_symbols=excluded_symbols
        )

        return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "AFHR_FLPR_YN": "N",
                "OFL_YN": "",
                "INQR_DVSN": "01",
                "UNPR_DVSN": "01",
                "FUND_STTL_ICLD_YN": "N",
                "FNCG_AMT_AUTO_RDPT_YN": "N",
                "PRCS_DVSN": "01",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_cash(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_cash(response=response, base_market=base_market)

        return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        if base_market == "KRW":
            params = {"FID_COND_MRKT_DIV_CODE": "J", "FID_INPUT_ISCD": symbol}
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_orderbook(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_orderbook(response=response, base_market=base_market)

        return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        if base_market == "KRW":
            params = {"PRDT_TYPE_CD": "300", "PDNO": symbol}
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_security_info(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_security(response=response, base_market=base_market)

        return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        if base_market == "KRW":
            params = {"FID_COND_MRKT_DIV_CODE": "J", "FID_INPUT_ISCD": symbol}
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_ticker_price(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_ticker(response=response, base_market=base_market)

        return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def fetch_historical_data_index(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        if time_frame.endswith("D"):
            param_code = "D"
        elif time_frame.endswith("W") or time_frame.endswith("w"):
            param_code = "W"
        elif time_frame.endswith("M"):
            param_code = "M"
        elif time_frame.endswith("Y"):
            param_code = "Y"
        else:
            assert ValueError(f"{time_frame} is not valid value")

        if start is None:
            start = self.now(base_market) - timedelta(days=50)
        if end is None:
            end = self.now(base_market)

        if base_market == "KRW":
            params = {
                "FID_COND_MRKT_DIV_CODE": "U",
                "FID_INPUT_ISCD": symbol,
                "FID_INPUT_DATE_1": start.strftime("%Y%m%d"),
                "FID_INPUT_DATE_2": end.strftime("%Y%m%d"),
                "FID_PERIOD_DIV_CODE": param_code,
            }
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_index_ohlcv(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtHistoricalDataResponse(header=common_header, response=common_response, info=None)

        parsed_response = self.parser.parse_historical_index_data(
            response=response, symbol=symbol, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtHistoricalDataResponse(
            header=common_header, response=common_response, info=parsed_response
        )

    @AsyncExchange.check_token
    async def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        if time_frame.endswith("D"):
            param_code = "D"
        elif time_frame.endswith("W") or time_frame.endswith("w"):
            param_code = "W"
        elif time_frame.endswith("M"):
            param_code = "M"
        elif time_frame.endswith("Y"):
            param_code = "Y"
        else:
            assert ValueError(f"{time_frame} is not valid value")

        if start is None:
            start = self.now(base_market) - timedelta(days=100)
        if end is None:
            end = self.now(base_market)

        if base_market == "KRW":
            params = {
                "FID_COND_MRKT_DIV_CODE": "J",
                "FID_INPUT_ISCD": symbol,
                "FID_INPUT_DATE_1": start.strftime("%Y%m%d"),
                "FID_INPUT_DATE_2": end.strftime("%Y%m%d"),
                "FID_PERIOD_DIV_CODE": param_code,
                "FID_ORG_ADJ_PRC": "0",
            }
        else:
            assert ValueError(f"{base_market} is not valid value")

        if time_frame.endswith("m"):
            common_header = self.create_common_header(request_params=params)
            response = await self.private_get_fetch_security_ohlcv_minute(self.extend(params))
        elif time_frame.endswith("D"):
            common_header = self.create_common_header(request_params=params)
            response = await self.private_get_fetch_security_ohlcv_day(self.extend(params))
        elif time_frame.endswith("W"):
            common_header = self.create_common_header(request_params=params)
            response = await self.private_get_fetch_security_ohlcv_week(self.extend(params))
        elif time_frame.endswith("M"):
            common_header = self.create_common_header(request_params=params)
            response = await self.private_get_fetch_security_ohlcv_month(self.extend(params))
        elif time_frame.endswith("Y"):
            common_header = self.create_common_header(request_params=params)
            response = await self.private_get_fetch_security_ohlcv_year(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtHistoricalDataResponse(header=common_header, response=common_response, info=None)

        parsed_response = self.parser.parse_historical_data(
            response=response, symbol=symbol, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtHistoricalDataResponse(
            header=common_header, response=common_response, info=parsed_response
        )

    @AsyncExchange.check_token
    async def fetch_deposit_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtDepositHistoryResponse:
        params = {"state": "ACCEPTED", "uuids": "", "txids": ""}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_deposit_history(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtWithdrawalHistoryResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_deposit_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtWithdrawalHistoryResponse(
            header=common_header, response=common_response, info=parsed_info
        )

    @AsyncExchange.check_token
    async def fetch_withdrawal_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        params = {"state": "done", "uuids": "", "txids": ""}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_withdrawal_history(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtWithdrawalHistoryResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_withdrawal_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtWithdrawalHistoryResponse(
            header=common_header, response=common_response, info=parsed_info
        )

    @AsyncExchange.check_token
    async def modify_order(
        self,
        acc_num: str,
        order_id: str,
        price: float,
        qty: float,
        *args,
        symbol: str | None = "",
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtModifyOrderResponse:
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "KRX_FWDG_ORD_ORGNO": "",
                "ORGN_ODNO": str(order_id),
                "RVSE_CNCL_DVSN_CD": "01",
                "ORD_DVSN": "00",
                "ORD_QTY": str(qty),
                "ORD_UNPR": str(price),
                "QTY_ALL_ORD_YN": "N",
            }
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)
        response = await self.private_post_send_modify_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtModifyOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_modify_order(response=response, base_market=base_market)

        return ksxt.models.KsxtModifyOrderResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def cancel_order(
        self, acc_num: str, order_id: str, symbol: str | None = "", qty: float = 0, *args, base_market: str = "KRW"
    ) -> ksxt.models.KsxtCancelOrderResponse:
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "KRX_FWDG_ORD_ORGNO": "",
                "ORGN_ODNO": str(order_id),
                "RVSE_CNCL_DVSN_CD": "02",
                "ORD_DVSN": "00",
                "ORD_QTY": str(qty),
                "ORD_UNPR": str(0),
                "QTY_ALL_ORD_YN": "N",
            }
        else:
            assert ValueError(f"{base_market} is not valid value")

        common_header = self.create_common_header(request_params=params)
        response = await self.private_post_send_cancel_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_cancel_order(response=response, base_market=base_market)

        return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=parsed_info)

    @AsyncExchange.check_token
    async def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong"] | Literal["EntryShort"] | Literal["ExitLong"] | Literal["ExitShort"],
        otype: Literal["limit"] | Literal["market"],
        price: float | None = 0,
        qty: float | None = 0,
        amount: float | None = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:
        if otype.lower() == "limit":
            order_dvsn = "00"
        elif otype.lower() == "market":
            order_dvsn = "01"
        params = {
            "CANO": acc_num[:8],
            "ACNT_PRDT_CD": acc_num[-2:],
            "PDNO": symbol,
            "ORD_DVSN": order_dvsn,
            "ORD_QTY": str(qty),  # string type 으로 설정
            "ORD_UNPR": str(price),  # string type 으로 설정
        }

        common_header = self.create_common_header(request_params=params)

        if ticket_type == "EntryLong":
            response = await self.private_post_send_order_entry(self.extend(params))
        elif ticket_type == "ExitLong":
            response = await self.private_post_send_order_exit(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_create_order(response=response, base_market=base_market)

        return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=parsed_info)

    def get_market_code_in_feeder(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            return ""
        elif base_market == "USD":
            if symbol.upper() == "ALL":
                return "NASD"

            response = self.fetch_security(symbol=symbol, base_market=base_market)
            return response["exchange"]
        else:
            return ""

    def get_market_code_in_broker(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            return ""
        elif base_market == "USD":
            if symbol.upper() == "ALL":
                return "NASD"

            response = self.fetch_security(symbol=symbol, base_market=base_market)
            exname = response["exchange"]
            if exname == "NYS":
                return "NYSE"
            elif exname == "NAS":
                return "NASD"
            elif exname == "AMS":
                return "AMEX"
            else:
                return ""
        else:
            return ""



entries = {
    "koreainvest": {
        "stock": {
            "private": {
                "get": [
                    "fetch_balance",
                    "fetch_cash",
                    "fetch_orderbook",
                    "fetch_security_info",
                    "fetch_ticker_price",
                    "fetch_index_ohlcv",
                    "fetch_security_ohlcv_day",
                    "fetch_security_ohlcv_minute",
                    "fetch_security_ohlcv_month",
                    "fetch_security_ohlcv_week",
                    "fetch_security_ohlcv_year",
                    "fetch_closed_order",
                    "fetch_opened_order",
                    "fetch_pnl",
                    "fetch_screener",
                    "fetch_screener_list",
                ],
                "post": [
                    "send_modify_order",
                    "send_cancel_order",
                    "send_order_entry",
                    "send_order_exit",
                    "send_order_loan_entry",
                    "send_order_loan_exit",
                ],
            },
            "public": {
                "post": [
                    "generate_token",
                    "revoke_token",
                ]
            },
        },
    },
    "websocket": {
        "stock": {
            "subscribe": {
                "ticker": ["subscribe_ticker"],
                "trade": ["subscribe_trade"],
                "orderbook": ["subscribe_orderbook"],
            }
        }
    },
}



from pydantic import BaseModel
from typing import List, Optional
from datetime import datetime

from ksxt.models.common import GeneralResponse


class OrderBase(BaseModel):
    datetime: datetime
    order_id: str
    symbol: str
    name: str
    price: float
    qty: float


class CreateOrderResponse(BaseModel):
    order_datetime: datetime
    order_id: str


class CancelOrderResponse(BaseModel):
    order_datetime: datetime
    order_id: str


class ModifyOrderResponse(BaseModel):
    order_datetime: datetime
    order_id: str


# Orders에 대한 구체적인 Response 타입
class KsxtCreateOrderResponse(GeneralResponse[CreateOrderResponse]):
    pass


class KsxtCancelOrderResponse(GeneralResponse[CancelOrderResponse]):
    pass


class KsxtModifyOrderResponse(GeneralResponse[ModifyOrderResponse]):
    pass



import hashlib
import json
import time
import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Literal, Optional
from urllib.parse import urlencode

import jwt

from ksxt.api.upbit import ImplicitAPI
from ksxt.async_.base.async_exchange import AsyncExchange
from ksxt.parser.upbit import UpbitParser
import ksxt.models


class Upbit(AsyncExchange, ImplicitAPI):
    def __init__(self, config: Dict = None):
        super().__init__(config, "upbit.toml")
        self.parser = UpbitParser()

    def safe_symbol(self, base_market: str, security: str) -> str:
        # If security already contains a hyphen, assume it's correctly formatted
        if "-" in security:
            return security

        # Otherwise, assume security is just the symbol and format it with base_market
        return f"{base_market}-{security}"

    def safe_security(self, symbol: str) -> str:
        if symbol.find("-") < 0:
            return symbol
        return symbol.split("-")[1]

    def sign(
        self,
        path,
        security_type,
        method_type,
        api_type: Any = "public",
        headers: Optional[Any] = None,
        body: Optional[Any] = None,
        params: Optional[Any] = None,
        config={},
    ):
        host_url = self.apis[self.type]["hostname"]
        destination = self.apis[self.type][security_type][path]["url"]
        version = self.apis[self.type]["version"]
        destination = self.implode_params(destination, params)

        url = host_url + "/" + version + "/" + destination

        if api_type == "private":
            payload = {
                "access_key": self.open_key,
                "nonce": str(uuid.uuid4()),
            }

            if params is not None and len(params) > 0:
                m = hashlib.sha512()
                m.update(urlencode(params).encode())
                payload.update({"query_hash": m.hexdigest(), "query_hash_alg": "SHA512"})

            jwt_token = jwt.encode(payload=payload, key=self.secret_key)
            authorization_token = f"Bearer {jwt_token}"

            if headers is None:
                headers = {}
                headers.update({"content-type": "application/json", "Authorization": authorization_token})

        if method_type.upper() == "POST":
            body = json.dumps(params)
            params = {}

        return {"url": url, "method": method_type, "headers": headers, "body": body, "params": params}

    def get_common_response(self, response):
        if "error" in response:
            return self.create_common_response(
                success="1",
                msg_code=self.safe_string(response["error"], "name"),
                msg=self.safe_string(response["error"], "message"),
                info=response,
            )

        return self.create_common_response(
            success="0",
            msg_code=self.safe_string(response, "name"),
            msg=self.safe_string(response, "message"),
            info=response,
        )

    async def fetch_markets(self, base_market: str = "KRW") -> ksxt.models.KsxtMarketResponse:
        params = {"isDetails": "True"}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_markets(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtMarketResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_markets(response=response, base_market=base_market)

        return ksxt.models.KsxtMarketResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_historical_data(
        self, symbol: str, time_frame: str, start: str | None = None, end: str | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        params = {"market": self.safe_symbol(base_market, symbol), "count": 200}

        if end:
            end_utc = end.astimezone(timezone.utc)
            params.update({"to": end_utc.strftime("%Y-%m-%d %H:%M:%S")})

        # TODO : time_frame 을 어떻게 고정시킬까? 우리는 분봉, 일봉, 주봉, 월봉 만 지원한다고 가정하면?
        if time_frame.endswith("m"):
            # TODO : parse number
            period = 1
            params.update({"unit": period})

            common_header = self.create_common_header(request_params=params)
            response = await self.public_get_fetch_security_ohlcv_minute(self.extend(params))

        elif time_frame.endswith("D"):
            common_header = self.create_common_header(request_params=params)
            response = await self.public_get_fetch_security_ohlcv_day(self.extend(params))

        elif time_frame.endswith("W"):
            common_header = self.create_common_header(request_params=params)
            response = await self.public_get_fetch_security_ohlcv_week(self.extend(params))

        elif time_frame.endswith("M"):
            common_header = self.create_common_header(request_params=params)
            response = await self.public_get_fetch_security_ohlcv_month(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtHistoricalDataResponse(header=common_header, response=common_response, info=None)

        parsed_response = self.parser.parse_historical_data(
            response=response, symbol=symbol, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtHistoricalDataResponse(
            header=common_header, response=common_response, info=parsed_response
        )

    async def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        params = {"markets": self.safe_symbol(base_market, symbol)}  # 다중 조회 시, 콤마로 구분 ex. 'KRW-BTC, BTC-ETH'

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_ticker_price(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_ticker(response=response, base_market=base_market)

        return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_tickers(self, symbols: List[str], base_market: str = "KRW") -> ksxt.models.KsxtTickersResponse:
        symbol = ",".join([self.safe_symbol(base_market, symbol) for symbol in symbols])
        params = {"markets": symbol}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_tickers_price(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtTickersResponse(header=common_header, response=common_response, info=[])

        # 데이터 파싱
        parsed_info = self.parser.parse_tickers(response, base_market)

        return ksxt.models.KsxtTickersResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        params = {"markets": self.safe_symbol(base_market, symbol), "level": 0}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_orderbook(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_orderbook(response=response, base_market=base_market)

        return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_orderbooks(
        self, symbols: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtMultiOrderBookResponse:
        symbol = ",".join([self.safe_symbol(base_market, symbol) for symbol in symbols])
        params = {"markets": symbol}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_orderbooks(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtMultiOrderBookResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_orderbooks(response, base_market)

        return ksxt.models.KsxtMultiOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_balance(
        self, acc_num: str, base_market: str = "KRW", excluded_symbols: Optional[list[str]] = None
    ) -> ksxt.models.KsxtBalanceResponse:
        params = {}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_balance(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_balance(
            response=response, base_market=base_market, excluded_symbols=excluded_symbols
        )

        return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        params = {}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_cash(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_cash(response=response, base_market=base_market)

        return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        params = {"market": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_security_info(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_security(response=response, base_market=base_market)

        return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_trade_fee(
        self, symbol: Optional[str] = "", base_market: str = "KRW"
    ) -> ksxt.models.KsxtTradeFeeResponse:
        params = {"market": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_trade_fee(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTradeFeeResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_trade_fee(response=response, base_market=base_market)
        return ksxt.models.KsxtTradeFeeResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_closed_order_detail(
        self, acc_num: str, order_ids: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtClosedOrderResponse:
        params = {
            "uuids": order_ids,
        }

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_closed_order_detail(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_closed_order_history(response, base_market)

        return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_open_order_detail(
        self, acc_num: str, order_ids: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtOpenOrderResponse:
        params = {
            "uuids": order_ids,
        }

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_opened_order_detail(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_open_order_history(response, base_market)

        return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_closed_order(
        self,
        acc_num: str,
        symbol: str = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtClosedOrderResponse:
        params = {"state": "done"}

        if symbol:
            params.update({"market": self.safe_symbol(base_market, symbol)})

        if start and end:
            duration = end - start
            if duration.days > 7:
                raise ValueError("The duration between start and end cannot exceed 7 days.")

        elif start:
            start_timestamp = int(start.timestamp() * 1000)
            params.update({"start_time": start_timestamp})
        elif end:
            end_timestamp = int(end.timestamp() * 1000)
            params.update({"end_time": end_timestamp})

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_closed_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_closed_order_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_open_order(
        self,
        acc_num: str,
        symbol: str = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtOpenOrderResponse:
        params = {"state": "wait"}

        if symbol:
            params.update({"market": self.safe_symbol(base_market, symbol)})

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_opened_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_open_order_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_withdrawal_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        params = {"currency": base_market, "state": "DONE"}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_withdrawal_history(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtWithdrawalHistoryResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_withdrawal_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtWithdrawalHistoryResponse(
            header=common_header, response=common_response, info=parsed_info
        )

    async def fetch_deposit_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtDepositHistoryResponse:
        params = {"currency": base_market, "state": "ACCEPTED"}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_deposit_history(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtDepositHistoryResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_deposit_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtDepositHistoryResponse(header=common_header, response=common_response, info=parsed_info)

    async def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong", "EntryShort", "ExitLong", "ExitShort"],
        otype: Literal["limit", "market"],
        price: Optional[float] = 0,
        qty: Optional[float] = 0,
        amount: Optional[float] = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:

        params = {
            "market": self.safe_symbol(base_market, symbol),
        }

        if ticket_type == "EntryLong":
            order_side = "bid"

            if otype == "limit":
                params.update({"side": order_side, "price": str(price), "volume": str(qty), "ord_type": "limit"})

                response = await self.private_post_send_order_entry(self.extend(params))
            elif otype == "market":
                params.update(
                    {
                        "side": order_side,
                        "price": amount,
                        "ord_type": "price",
                    }
                )

                response = await self.private_post_send_order_entry_market(self.extend(params))
        elif ticket_type == "ExitLong":
            order_side = "ask"
            params = {
                "market": self.safe_symbol(base_market, symbol),
                "side": order_side,
            }

            if otype == "limit":
                params.update({"price": str(price), "volume": str(qty), "ord_type": "limit"})

                response = await self.private_post_send_order_exit(self.extend(params))

            elif otype == "market":
                params.update({"ord_type": "market", "volume": str(qty)})

                response = await self.private_post_send_order_exit_market(self.extend(params))
        else:
            raise ValueError(
                f"Unsupported ticket_type '{ticket_type}'. {__class__.__name__} supports only 'EntryLong' and 'ExitLong'."
            )

        common_header = self.create_common_header(request_params=params)
        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_create_order(response=response, base_market=base_market)

        return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def cancel_order(
        self,
        acc_num: str,
        order_id: str,
        symbol: str | None = "",
        qty: float = 0,
        *args,
        base_market: str = "KRW",
        **kwargs,
    ) -> ksxt.models.KsxtCancelOrderResponse:
        params = {"uuid": order_id}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_delete_send_cancel_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_cancel_order(response=response, base_market=base_market)

        return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=parsed_info)



from ksxt.api import Entry


class ImplicitAPI:

    # stock API methods

    # private methods

    # get requests
    private_get_fetch_balance = Entry("stock", "fetch_balance", {})
    private_get_fetch_cash = Entry("stock", "fetch_cash", {})
    private_get_fetch_orderbook = Entry("stock", "fetch_orderbook", {})
    private_get_fetch_security_info = Entry("stock", "fetch_security_info", {})
    private_get_fetch_ticker_price = Entry("stock", "fetch_ticker_price", {})
    private_get_fetch_index_ohlcv = Entry("stock", "fetch_index_ohlcv", {})
    private_get_fetch_security_ohlcv_day = Entry("stock", "fetch_security_ohlcv_day", {})
    private_get_fetch_security_ohlcv_minute = Entry("stock", "fetch_security_ohlcv_minute", {})
    private_get_fetch_security_ohlcv_month = Entry("stock", "fetch_security_ohlcv_month", {})
    private_get_fetch_security_ohlcv_week = Entry("stock", "fetch_security_ohlcv_week", {})
    private_get_fetch_security_ohlcv_year = Entry("stock", "fetch_security_ohlcv_year", {})
    # private_get_fetch_closed_order          = Entry('stock', 'fetch_closed_order', {})
    # private_get_fetch_opened_order          = Entry('stock', 'fetch_opened_order', {})
    # private_get_fetch_pnl                   = Entry('stock', 'fetch_pnl', {})
    # private_get_fetch_screener              = Entry('stock', 'fetch_screener', {})
    # private_get_fetch_screener_list         = Entry('stock', 'fetch_screener_list', {})

    # post requests
    private_post_send_modify_order = Entry("stock", "send_modify_order", {})
    private_post_send_cancel_order = Entry("stock", "send_cancel_order", {})
    private_post_send_order_entry = Entry("stock", "send_order_entry", {})
    private_post_send_order_exit = Entry("stock", "send_order_exit", {})
    private_post_send_order_loan_entry = Entry("stock", "send_order_loan_entry", {})
    private_post_send_order_loan_exit = Entry("stock", "send_order_loan_exit", {})

    # public methods

    # post requests
    public_post_generate_token = Entry("token", "generate_token", {})
    public_post_revoke_token = Entry("token", "revoke_token", {})



import json
import logging
import tomllib
import toml
import os
from pathlib import Path

from typing import Any, Dict, List, Literal

from datetime import datetime, UTC
import pytz
import time

from ksxt.base.rate_limiter import (
    RateLimiterContext,
    RequestRateLimiter,
    TimeBasedRateLimiter,
)
from ksxt.base.errors import NotSupportedError
from ksxt.base.exchange import Exchange
from ksxt.config import CONFIG_DIR
import ksxt.models


class RestExchange(Exchange):
    required_credentials = {
        "open_key": True,
        "secret_key": True,
        "uid": False,
        "login": False,
        "password": False,
        "token": False,
    }

    headers = None
    token = None
    token_expired = None
    type = "rest"

    rate_limit = 2000  # milliseconds = seconds * 1000
    lastRestRequestTimestamp = 0
    timezone = pytz.timezone("UTC")

    def __init__(self, config: Dict = None, filename: str = None) -> None:
        super().__init__()

        self.rate_limiters: Dict[str, RateLimiterContext] = {}

        self.headers = dict() if self.headers is None else self.headers

        if config is None:
            config = {}

        # settings = self.deep_extend(self.describe(), config)
        # Exchange.set_attr(self, settings)

        settings = self.deep_extend(config)
        Exchange.set_attr(self, settings)

        apis = self._get_api_from_file(filename)
        Exchange.set_attr(self, apis)

    def _get_api_from_file(self, filename: str) -> Dict[str, Any]:
        if filename is None:
            raise ValueError("Configuration filename cannot be None.")

        config_path = os.path.join(CONFIG_DIR, filename)

        if Path(filename).suffix != ".toml":
            raise ValueError(f"Unsupported file format: {Path(filename).suffix}")

        with open(config_path, mode="rb") as f:
            config = tomllib.load(f)

        self._setup_rate_limiters(config)
        return config

    def _setup_rate_limiters(self, config: Dict[str, Any]):
        """RateLimiter 설정을 구성하는 메서드"""
        default_rate_limit = config.get("rate_limit", 1000)  # 최상위 rate_limit 값 가져오기

        apis = config.get("apis", {}).get("rest", {})

        def traverse_apis(api_section, section_name="", self=self):
            for key, value in api_section.items():
                if not isinstance(value, dict):
                    continue

                if "url" in value:  # "url"이 있는 경우, API 엔드포인트로 간주
                    rate_limit = value.get("rate_limit", default_rate_limit)
                    rate_limit_period = value.get("rate_limit_period", 1.0)

                    # 전략 패턴을 이용한 RateLimiter 생성
                    if rate_limit_period > 1.0:
                        strategy = TimeBasedRateLimiter(period=rate_limit_period)
                    else:
                        strategy = RequestRateLimiter(max_requests=rate_limit, period=1.0)

                    # API 이름을 구성
                    api_name = f"{section_name}.{key}" if section_name else key
                    self.rate_limiters[api_name] = RateLimiterContext(strategy)
                else:
                    # value가 딕셔너리지만 "url"이 없는 경우, 하위 섹션으로 간주하고 재귀 호출
                    traverse_apis(value, f"{section_name}.{key}" if section_name else key)

        traverse_apis(apis)

    def check_token(func):
        def wrapper(self, *args, **kwargs):
            if not self.is_valid_token():
                self.create_token()

            return func(self, *args, **kwargs)

        return wrapper

    def is_valid_token(self) -> bool:
        self.load_token()
        if self.token_expired is None:
            return False

        return datetime.now(self.timezone) <= self.timezone.localize(self.token_expired)

    def read_token(self):
        try:
            token_path = os.path.join(CONFIG_DIR, "token.toml")
            with open(token_path, mode="r") as f:
                c = toml.load(f)
            return c
        except Exception as e:
            return {}

    def load_token(self):
        try:
            data = self.read_token()

            if self.open_key in data.keys():
                self.token = data[self.open_key]["token"]
                self.token_expired = data[self.open_key]["expired"]
        except Exception as e:
            pass

    def save_token(self, open_key, token, expired):
        try:
            logging.info("save token")
            logging.info(open_key)
            logging.info(token)
            logging.info(expired)

            data = self.read_token()
            if open_key not in data:
                data[open_key] = {}
            data[open_key]["token"] = token
            data[open_key]["expired"] = expired

            token_path = os.path.join(CONFIG_DIR, "token.toml")
            with open(token_path, "w") as file:
                toml.dump(data, file)

            self.load_token()
        except Exception as e:
            pass

    def create_token(self):
        pass

    def prepare_request_headers(self, headers=None):
        headers = headers or {}

        if self.session:
            headers.update(self.session.headers)

        self.headers.update(headers)

        headers.update({"content-type": "application/json"})
        # headers.update({'appKey':self.open_key})
        # headers.update({'appsecret':self.secret_key})

        return headers

    def throttle(self, cost=None):
        now = float(self.milliseconds())
        elapsed = now - self.lastRestRequestTimestamp
        cost = 1 if cost is None else cost
        sleep_time = self.rate_limit * cost
        if elapsed < sleep_time:
            delay = sleep_time - elapsed
            time.sleep(delay / 1000.0)

    def fetch(self, url, method="GET", headers=None, body=None, params=None):
        request_headers = headers
        request_body = body
        request_params = params

        self.session.cookies.clear()

        http_response = None
        http_status_code = None
        http_status_text = None
        json_response = None

        try:
            response = self.session.request(
                method=method,
                url=url,
                headers=request_headers,
                data=request_body,
                params=request_params,
                timeout=int(self.timeout / 1000),
            )

            response.encoding = "utf-8"

            headers = response.headers
            http_status_code = response.status_code
            http_status_text = response.reason
            http_response = response.text.strip()
            json_response = self.parse_json(http_response)

        except TimeoutError as e:
            details = " ".join([self.id, method, url])
            raise TimeoutError(details) from e

        return json_response

    def is_activate(self, path, security_type) -> bool:
        return self.apis[self.type][security_type][path]["activate"]

    def sign(
        self,
        path,
        security_type,
        method_type,
        api_type: Any = "public",
        headers: Any | None = None,
        body: Any | None = None,
        params: Any | None = None,
        config={},
    ):
        pass

    def fetch2(self, path, security_type, params={}, headers: Any | None = None, body: Any | None = None, config={}):
        is_activate = self.is_activate(path=path, security_type=security_type)
        if not is_activate:
            return {
                "response": {
                    # 성공 실패 여부
                    "success": "-1",
                    # 응답코드
                    "code": "fail",
                    # 응답메세지
                    "message": f"지원하지 않는 함수({path}) 입니다.",
                }
            }

        # if self.enableRateLimit:
        #     cost = self.calculate_rate_limiter_cost(api, method, path, params, config)
        #     self.throttle(cost)

        # self.lastRestRequestTimestamp = self.milliseconds()

        method_type = self.apis[self.type][security_type][path]["method"]
        api_type = self.apis[self.type][security_type][path]["api"]
        request = self.sign(path, security_type, method_type, api_type, headers, body, params, config)
        return self.fetch(request["url"], request["method"], request["headers"], request["body"], request["params"])

    def request(self, path, security_type, params={}, headers: Any | None = None, body: Any | None = None, config={}):
        return self.fetch2(path, security_type, params, headers, body, config)

    # region base method
    def create_token(self) -> ksxt.models.KsxtTokenResponse:
        raise NotSupportedError(f"{self.id} {self.create_token.__qualname__}() is not supported yet.")

    @check_token
    def fetch_markets(self, market_name: str) -> ksxt.models.KsxtMarketResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_markets.__qualname__}() is not supported yet.")

    @check_token
    def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_security.__qualname__}() is not supported yet.")

    @check_token
    def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_ticker.__qualname__}() is not supported yet.")

    @check_token
    def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_orderbook.__qualname__}() is not supported yet.")

    @check_token
    def fetch_historical_data_index(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_historical_data_index.__qualname__}() is not supported yet.")

    @check_token
    def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_historical_data.__qualname__}() is not supported yet.")

    @check_token
    def fetch_is_holiday(self, dt: datetime, base_market: str = "KRW") -> ksxt.models.KsxtHolidayResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_is_holiday.__qualname__}() is not supported yet.")

    @check_token
    def fetch_user_info(self, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_user_info.__qualname__}() is not supported yet.")

    @check_token
    def fetch_balance(
        self, acc_num: str, base_market: str = "KRW", excluded_symbols: list[str] | None = None
    ) -> ksxt.models.KsxtBalanceResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_balance.__qualname__}() is not supported yet.")

    @check_token
    def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_cash.__qualname__}() is not supported yet.")

    @check_token
    def fetch_trade_fee(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTradeFeeResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_trade_fee.__qualname__}() is not supported yet.")

    @check_token
    def fetch_screener_list(self, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_screener_list.__qualname__}() is not supported yet.")

    @check_token
    def fetch_screener(self, screen_id: str, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_screener.__qualname__}() is not supported yet.")

    @check_token
    def fetch_deposit_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtDepositHistoryResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_deposit_history.__qualname__}() is not supported yet.")

    @check_token
    def fetch_withdrawal_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_withdrawal_history.__qualname__}() is not supported yet.")

    @check_token
    def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong", "EntryShort", "ExitLong", "ExitShort"],
        otype: Literal["limit", "market"],
        price: float | None = 0,
        qty: float | None = 0,
        amount: float | None = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:
        raise NotSupportedError(f"{self.id} {self.create_order.__qualname__}() is not supported yet.")

    @check_token
    def cancel_order(
        self, acc_num: str, order_id: str, symbol: str | None = "", qty: float = 0, *args, base_market: str = "KRW"
    ) -> ksxt.models.KsxtCancelOrderResponse:
        raise NotSupportedError(f"{self.id} {self.cancel_order.__qualname__}() is not supported yet.")

    @check_token
    def modify_order(
        self,
        acc_num: str,
        order_id: str,
        price: float,
        qty: float,
        *args,
        symbol: str | None = "",
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtModifyOrderResponse:
        raise NotSupportedError(f"{self.id} {self.modify_order.__qualname__}() is not supported yet.")

    @check_token
    def fetch_open_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtOpenOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_open_order.__qualname__}() is not supported yet.")

    @check_token
    def fetch_closed_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtClosedOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_closed_order.__qualname__}() is not supported yet.")

    @check_token
    def fetch_open_order_detail(
        self, acc_num: str, order_ids: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtOpenOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_open_order_detail.__qualname__}() is not supported yet.")

    @check_token
    def fetch_closed_order_detail(
        self, acc_num: str, order_ids: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtOpenOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_closed_order_detail.__qualname__}() is not supported yet.")

    @check_token
    def reserve_order(
        self, acc_num: str, symbol: str, price: float, qty: float, target_date: str, base_market: str = "KRW"
    ):
        raise NotSupportedError(f"{self.id} {self.reserve_order.__qualname__}() is not supported yet.")

    # endregion base method



from datetime import datetime
from typing import Dict, List, Optional
import inspect

from ksxt.models.balance import BalanceData, BalanceInfo
from ksxt.models.cash import CashInfo
from ksxt.models.historical import HistoricalDataInfo, OHLCVData
from ksxt.models.market import HolidayHistory, HolidayInfo, MarketInfo, SecurityData, TradeFeeInfo
from ksxt.models.order import CancelOrderResponse, CreateOrderResponse, ModifyOrderResponse
from ksxt.models.orderbook import MultiSymbolOrderBookInfos, OrderBookData, OrderBookInfo
from ksxt.models.ticker import MultiSymbolTickerInfo, TickerInfo
from ksxt.models.transaction import (
    ClosedOrderHistory,
    ClosedOrderInfo,
    DepositHistory,
    OpenedOrderHistory,
    OpenedOrderInfo,
    TransactionInfo,
    WithdrawalHistory,
)


class BaseParser:
    def _raise_not_implemented_error(self):
        caller = inspect.stack()[1].function
        raise NotImplementedError(f"{self.__class__.__name__} called {caller}")

    def parse_markets(self, response: List[Dict], base_market: str = "KRW") -> MarketInfo:
        self._raise_not_implemented_error()

    def _parse_market(self, market: Dict) -> SecurityData:
        self._raise_not_implemented_error()

    def parse_historical_data(
        self,
        response: List[Dict],
        symbol: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> HistoricalDataInfo:
        self._raise_not_implemented_error()

    def _parse_ohlcva(self, ohlcva: dict, base_market: str = "KRW") -> OHLCVData:
        self._raise_not_implemented_error()

    def parse_historical_index_data(self, response: dict, symbol: str, base_market: str = "KRW") -> HistoricalDataInfo:
        self._raise_not_implemented_error()

    def _parse_index_ohlcva(self, ohlcva: dict, base_market: str = "KRW") -> OHLCVData:
        self._raise_not_implemented_error()

    def parse_ticker(self, response: List[Dict], base_market: str = "KRW") -> TickerInfo:
        self._raise_not_implemented_error()

    def parse_tickers(self, response: List[Dict], base_market: str = "KRW") -> MultiSymbolTickerInfo:
        self._raise_not_implemented_error()

    def _parse_ticker(self, response: dict, base_market: str) -> TickerInfo:
        self._raise_not_implemented_error()

    def parse_orderbook(self, response: List[Dict], base_market: str = "KRW") -> OrderBookInfo:
        self._raise_not_implemented_error()

    def parse_orderbooks(self, response: dict, base_market: str) -> MultiSymbolOrderBookInfos:
        self._raise_not_implemented_error()

    def _parse_orderbook(
        self, orderbook: dict, index: int, base_market: str = "KRW"
    ) -> tuple[OrderBookData, OrderBookData]:
        self._raise_not_implemented_error()

    def parse_balance(
        self, response: List[Dict], base_market: str = "KRW", excluded_symbols: Optional[list[str]] = None
    ) -> BalanceInfo:
        self._raise_not_implemented_error()

    def _parse_balance(self, data: Dict, base_market: str = "KRW") -> BalanceData:
        self._raise_not_implemented_error()

    def parse_cash(self, response: List[Dict], base_market: str = "KRW") -> CashInfo:
        self._raise_not_implemented_error()

    def parse_security(self, response: dict, base_market: str = "KRW") -> SecurityData:
        self._raise_not_implemented_error()

    def parse_trade_fee(self, response: dict, base_market: str = "KRW") -> TradeFeeInfo:
        self._raise_not_implemented_error()

    def parse_open_order_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> OpenedOrderHistory:
        self._raise_not_implemented_error()

    def _parse_open_order_info(self, order: dict, base_market: str = "KRW") -> OpenedOrderInfo:
        self._raise_not_implemented_error()

    def parse_closed_order_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ClosedOrderHistory:
        self._raise_not_implemented_error()

    def _parse_closed_order_info(self, order: dict, base_market: str = "KRW") -> ClosedOrderInfo:
        self._raise_not_implemented_error()

    def parse_modify_order(self, response: dict, base_market: str = "KRW") -> ModifyOrderResponse:
        self._raise_not_implemented_error()

    def parse_cancel_order(self, response: dict, base_market: str = "KRW") -> CancelOrderResponse:
        self._raise_not_implemented_error()

    def parse_create_order(self, response: dict, base_market: str = "KRW") -> CreateOrderResponse:
        self._raise_not_implemented_error()

    def parse_withdrawal_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> WithdrawalHistory:
        self._raise_not_implemented_error()

    def parse_deposit_history(
        self, response: List[Dict], start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> DepositHistory:
        self._raise_not_implemented_error()

    def _parse_transaction_history_item(self, item: Dict, base_market: str = "KRW") -> TransactionInfo:
        self._raise_not_implemented_error()

    def parse_is_holiday(self, response: List[Dict], base_market: str = "KRW") -> HolidayHistory:
        self._raise_not_implemented_error()

    def _parse_is_holiday(self, item: Dict, base_market: str = "KRW") -> HolidayInfo:
        self._raise_not_implemented_error()



from typing import Optional
from pydantic import BaseModel

from ksxt.models.common import GeneralResponse


class ErrorInfo(BaseModel):
    error_code: Optional[str] = None
    error_message: Optional[str] = None


class KsxtErrorResponse(GeneralResponse[ErrorInfo]):
    pass



from datetime import datetime
from typing import Generic, Iterable, Iterator, Literal, TypeVar, get_args
from ksxt.market.base import MarketItemBase, MarketBase
from ksxt.market.db import Base


SCRT_GRP_CLS_CODES = {
    "ST": "주권",
    "MF": "증권투자회사",
    "RT": "부동산투자회사",
    "SC": "선박투자회사",
    "IF": "사회간접자본투융자회사",
    "DR": "주식예탁증서",
    "EW": "ELW",
    "EF": "ETF",
    "SW": "신주인수권증권",
    "SR": "신주인수권증서",
    "BC": "수익증권",
    "FE": "해외ETF",
    "FS": "외국주권",
}

AVLS_SCAL_CLS_CODES = {
    "0": "제외",
    "1": "대",
    "2": "중",
    "3": "소",
}

ETP_PROD_CLS_CODES = {
    "0": "해당없음",
    "1": "투자회사형",
    "2": "수익증권형",
    "3": "ETN",
    "4": "손실제한ETN",
}

SHORT_OVER_CLS_CODES = {
    "0": "해당없음",
    "1": "지정예고",
    "2": "지정",
    "3": "지정연장(해제연기)",
}

MRKT_ALRM_CLS_CODES = {
    "00": "해당없음",
    "01": "투자주의",
    "02": "투자경고",
    "03": "투자위험",
}

FLNG_CLS_CODES = {
    "00": "해당사항없음",
    "01": "권리락",
    "02": "배당락",
    "03": "분배락",
    "04": "권배락",
    "05": "중간배당락",
    "06": "권리중간배당락",
    "99": "기타",
}

FCAM_MOD_CLS_CODES = {
    "00": "해당없음",
    "01": "액면분할",
    "02": "액면병합",
    "99": "기타",
}

ICIC_CLS_CODES = {
    "00": "해당없음",
    "01": "유상증자",
    "02": "무상증자",
    "03": "유무상증자",
    "99": "기타",
}

PRST_CLS_CODES = {
    "0": "해당없음(보통주)",
    "1": "구형우선주",
    "2": "신형우선주",
}


class KrxStockItem(MarketItemBase):
    """국내 주식 정보"""

    excd: str
    """거래소 정보"""
    mksc_shrn_iscd: str
    """단축코드"""
    stnd_iscd: str
    """표준코드"""
    hts_kor_isnm: str
    """한글종목명"""
    scrt_grp_cls_code: Literal["ST", "MF", "RT", "SC", "IF", "DR", "EW", "EF", "SW", "SR", "BC", "FE", "FS"]
    """증권그룹구분코드"""
    curr: str
    """화폐 정보"""
    avls_scal_cls_code: Literal["0", "1", "2", "3"]
    """시가총액 규모 구분 코드 유가"""
    bstp_larg_div_code: str
    """지수 업종 대분류 코드"""
    bstp_medm_div_code: str
    """지수 업종 중분류 코드"""
    bstp_smal_div_code: str
    """지수 업종 소분류 코드"""
    low_current_yn: bool
    """저유동성종목 여부"""
    krx_issu_yn: bool
    """KRX 종목 여부"""
    etp_prod_cls_code: Literal["0", "1", "2", "3"]
    """ETP 상품구분코드"""
    krx100_issu_yn: bool
    """KRX100 종목 여부"""
    krx_car_yn: bool
    """KRX 자동차 여부"""
    krx_smcn_yn: bool
    """KRX 반도체 여부"""
    krx_bio_yn: bool
    """KRX 바이오 여부"""
    krx_bank_yn: bool
    """KRX 은행 여부"""
    etpr_undt_objt_co_yn: bool
    """기업인수목적회사여부"""
    krx_enrg_chms_yn: bool
    """KRX 에너지 화학 여부"""
    krx_stel_yn: bool
    """KRX 철강 여부"""
    short_over_cls_code: Literal["0", "1", "2", "3"]
    """단기과열종목구분코드"""
    krx_medi_cmnc_yn: bool
    """KRX 미디어 통신 여부"""
    krx_cnst_yn: bool
    """KRX 건설 여부"""
    krx_scrt_yn: bool
    """KRX 증권 구분"""
    krx_ship_yn: bool
    """KRX 선박 구분"""
    krx_insu_yn: bool
    """KRX섹터지수 보험여부"""
    krx_trnp_yn: bool
    """KRX섹터지수 운송여부"""
    stck_sdpr: int
    """주식 기준가"""
    frml_mrkt_deal_qty_unit: int
    """정규 시장 매매 수량 단위"""
    ovtm_mrkt_deal_qty_unit: int
    """시간외 시장 매매 수량"""
    trht_yn: bool
    """거래정지 여부"""
    sltr_yn: bool
    """정리매매 여부"""
    mang_issu_yn: bool
    """관리 종목 여부"""
    mrkt_alrm_cls_code: Literal["00", "01", "02", "03"]
    """시장 경고 구분 코드"""
    mrkt_alrm_risk_adnt_yn: bool
    """시장 경고위험 예고 여부"""
    insn_pbnt_yn: bool
    """불성실 공시 여부"""
    byps_lstn_yn: bool
    """우회 상장 여부"""
    flng_cls_code: Literal["00", "01", "02", "03", "04", "05", "06", "99"]
    """락구분 코드"""
    fcam_mod_cls_code: Literal["00", "01", "02", "99"]
    """액면가 변경 구분 코드"""
    icic_cls_code: Literal["00", "01", "02", "03", "99"]
    """증자 구분 코드"""
    marg_rate: int
    """증거금 비율"""
    crdt_able: bool
    """신용주문 가능 여부"""
    crdt_days: int
    """신용기간"""
    prdy_vol: int
    """전일 거래량"""
    stck_fcam: int
    """주식 액면가"""
    stck_lstn_date: datetime
    """주식 상장 일자"""
    lstn_stcn: int
    """상장 주수(천)"""
    cpfn: int
    """자본금"""
    stac_month: int
    """결산 월"""
    po_prc: int
    """공모 가격"""
    prst_cls_code: Literal["0", "1", "2"]
    """우선주 구분 코드"""
    ssts_hot_yn: bool
    """공매도과열종목여부"""
    stange_runup_yn: bool
    """이상급등종목여부"""
    krx300_issu_yn: bool
    """KRX300 종목 여부"""
    sale_account: int
    """매출액"""
    bsop_prfi: int
    """영업이익"""
    op_prfi: int
    """경상이익"""
    thtr_ntin: int
    """당기순이익"""
    roe: float
    """ROE(자기자본이익률)"""
    base_date: datetime
    """기준년월"""
    prdy_avls_scal: int
    """전일기준 시가총액 (억)"""
    grp_code: str
    """그룹사 코드"""
    co_crdt_limt_over_yn: bool
    """회사신용한도초과여부"""
    secu_lend_able_yn: bool
    """담보대출가능여부"""
    stln_able_yn: bool
    """대주가능여부"""

    @property
    def scrt_grp_cls_name(self) -> str:
        return SCRT_GRP_CLS_CODES.get(self.scrt_grp_cls_code, None)

    @property
    def avls_scal_cls_name(self) -> str:
        return AVLS_SCAL_CLS_CODES.get(self.avls_scal_cls_code, None)

    @property
    def etp_prod_cls_name(self) -> str:
        return ETP_PROD_CLS_CODES.get(self.etp_prod_cls_code, None)

    @property
    def short_over_cls_name(self) -> str:
        return SHORT_OVER_CLS_CODES.get(self.short_over_cls_code, None)

    @property
    def mrkt_alrm_cls_name(self) -> str:
        return MRKT_ALRM_CLS_CODES.get(self.mrkt_alrm_cls_code, None)

    @property
    def flng_cls_name(self) -> str:
        return FLNG_CLS_CODES.get(self.flng_cls_code, None)

    @property
    def fcam_mod_cls_name(self) -> str:
        return FCAM_MOD_CLS_CODES.get(self.fcam_mod_cls_code, None)

    @property
    def icic_cls_name(self) -> str:
        return ICIC_CLS_CODES.get(self.icic_cls_code, None)

    @property
    def prst_cls_name(self) -> str:
        return PRST_CLS_CODES.get(self.prst_cls_code, None)

    def __init__(self, data: str):
        super().__init__(data)


t_item = TypeVar("t_item", bound=KrxStockItem)
t_dbitem = TypeVar("t_dbitem", bound=Base)


class KrxStockMarket(Generic[t_item, t_dbitem], MarketBase[t_item, t_dbitem]):
    def search(self, name: str, limit: int = 50) -> Iterable[t_item]:
        """종목을 검색합니다."""
        _, db_type = get_args(self.__orig_bases__[0])  # type: ignore
        return self._search(db_type.hts_kor_isnm, name, limit)  # type: ignore

    def items(self, offset: int = 0, limit: int = 100) -> Iterable[t_item]:
        """종목을 가져옵니다."""
        return super().items(offset, limit)

    def all(self) -> Iterator[t_item]:
        """모든 종목을 가져옵니다."""
        return super().all()  # type: ignore

    def __getitem__(self, code: str) -> t_item:
        """종목을 가져옵니다."""
        _, db_type = get_args(self.__orig_bases__[0])  # type: ignore
        return self._get(db_type.mksc_shrn_iscd, code)



from abc import ABC, abstractmethod
import asyncio
import time


class RateLimiterStrategy(ABC):
    @abstractmethod
    async def async_acquire(self):
        pass

    @abstractmethod
    def acquire(self):
        pass

    @abstractmethod
    def release(self):
        pass


class RateLimiterContext:
    def __init__(self, strategy: RateLimiterStrategy):
        self._strategy = strategy

    async def async_acquire(self):
        await self._strategy.async_acquire()

    def acquire(self):
        self._strategy.acquire()

    def release(self):
        self._strategy.release()


class RequestRateLimiter(RateLimiterStrategy):
    def __init__(self, max_requests: int, period: float = 1.0):
        self.max_requests = max_requests
        self.period = period
        self.semaphore = asyncio.BoundedSemaphore(max_requests)
        self.last_reset_time = time.time()

    async def async_acquire(self):
        current_time = time.time()
        if current_time - self.last_reset_time > self.period:
            self.semaphore = asyncio.BoundedSemaphore(self.max_requests)
            self.last_reset_time = current_time

        await self.semaphore.acquire()

    def acquire(self):
        current_time = time.time()
        if current_time - self.last_reset_time > self.period:
            # 동기식 세마포어는 지원되지 않으므로, 대신에 현재 상황을 조정합니다.
            self.last_reset_time = current_time

        # 세마포어가 동기식으로 동작하지 않으므로, 대신 제한 시간을 체크하여 동기식으로 제한
        while self.semaphore._value <= 0:  # 내부 값이 0 이하라면, 대기
            if time.time() - self.last_reset_time > self.period:
                self.semaphore = asyncio.BoundedSemaphore(self.max_requests)
                break
            time.sleep(0.01)  # 동기식 대기

        self.semaphore._value -= 1  # 수동으로 세마포어 값을 감소시킴

    def release(self):
        self.semaphore.release()


class TimeBasedRateLimiter(RateLimiterStrategy):
    def __init__(self, period: float):
        self.period = period
        self.last_request_time = None

    async def async_acquire(self):
        current_time = time.time()
        if self.last_request_time and (current_time - self.last_request_time) < self.period:
            raise ValueError("요청이 너무 자주 발생했습니다. 잠시 후에 다시 시도해주세요.")

        self.last_request_time = current_time

    def acquire(self):
        current_time = time.time()
        if self.last_request_time and (current_time - self.last_request_time) < self.period:
            raise ValueError("요청이 너무 자주 발생했습니다. 잠시 후에 다시 시도해주세요.")

        self.last_request_time = current_time

    def release(self):
        pass  # 이 클래스에서는 release가 필요하지 않음



import base64
import hashlib
import hmac
import json
import time
from datetime import datetime, timezone
from typing import Any, Dict, List, Literal, Optional
from urllib.parse import urlencode
import uuid

import jwt

from ksxt.api.bithumb import ImplicitAPI
from ksxt.async_.base.async_exchange import AsyncExchange
from ksxt.parser.bithumb import BithumbParser

import ksxt.models


class Bithumb(AsyncExchange, ImplicitAPI):
    def __init__(self, config: Dict = None):
        super().__init__(config, "bithumb.toml")
        self.parser = BithumbParser()

    def safe_symbol(self, base_market: str, security: str) -> str:
        return f"{base_market}-{security}"

    def sign(
        self,
        path,
        security_type,
        method_type,
        api_type: Any = "public",
        headers: Optional[Any] = None,
        body: Optional[Any] = None,
        params: Optional[Any] = None,
        config={},
    ):
        host_url = self.apis[self.type]["hostname"]
        destination = self.apis[self.type][security_type][path]["url"]
        version = self.apis[self.type]["version"]
        destination = self.implode_params(destination, params)

        url = host_url + "/" + version + "/" + destination

        if api_type == "private":
            payload = {
                "access_key": self.open_key,
                "nonce": str(uuid.uuid4()),
                "timestamp": round(self.milliseconds()),
            }

            if params is not None and len(params) > 0:
                query = urlencode(params).encode()

                hash = hashlib.sha512()
                hash.update(query)
                query_hash = hash.hexdigest()

                payload.update({"query_hash": query_hash, "query_hash_alg": "SHA512"})

            jwt_token = jwt.encode(payload=payload, key=self.secret_key)
            authorization_token = f"Bearer {jwt_token}"

            if headers is None:
                headers = {}
                headers.update(
                    {"Content-Type": "application/json; charset=utf-8", "Authorization": authorization_token}
                )

        if method_type.upper() == "POST":
            body = json.dumps(params)
            params = {}

        return {"url": url, "method": method_type, "headers": headers, "body": body, "params": params}

    def get_common_response(self, response):
        if "error" in response:
            return self.create_common_response(
                success="1",
                msg_code=self.safe_string(response["error"], "name"),
                msg=self.safe_string(response["error"], "message"),
                info=response,
            )

        return self.create_common_response(
            success="0",
            msg_code=self.safe_string(response, "name"),
            msg=self.safe_string(response, "message"),
            info=response,
        )

    async def fetch_markets(self, market_name: str = "KRW") -> ksxt.models.KsxtMarketResponse:
        params = {
            # 유의종목 필드과 같은 상세 정보 노출 여부(선택 파라미터)
            "isDetails": "True"
        }

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_markets(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtMarketResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_markets(response=response, base_market=market_name)
        return ksxt.models.KsxtMarketResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        params = {"market": self.safe_symbol(base_market=base_market, security=symbol), "count": 200}

        if end:
            end_utc = end.astimezone(timezone.utc)
            params.update({"to": end_utc.strftime("%Y-%m-%d %H:%M:%S")})

        common_header = self.create_common_header(request_params=params)

        # TODO : time_frame 을 어떻게 고정시킬까? 우리는 분봉, 일봉, 주봉, 월봉 만 지원한다고 가정하면?
        if time_frame.endswith("m"):
            # TODO : parse number
            period = 1
            params.update({"unit": period})

            response = await self.public_get_fetch_security_ohlcv_minute(self.extend(params))

        elif time_frame.endswith("D"):
            response = await self.public_get_fetch_security_ohlcv_day(self.extend(params))

        elif time_frame.endswith("W"):
            response = await self.public_get_fetch_security_ohlcv_week(self.extend(params))

        elif time_frame.endswith("M"):
            response = await self.public_get_fetch_security_ohlcv_month(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtHistoricalDataResponse(header=common_header, response=common_response, info=None)

        parsed_response = self.parser.parse_historical_data(
            response=response, symbol=symbol, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtHistoricalDataResponse(
            header=common_header, response=common_response, info=parsed_response
        )

    async def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        params = {"markets": self.safe_symbol(base_market=base_market, security=symbol)}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_ticker_price(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=[])

        # 데이터 파싱
        parsed_info = self.parser.parse_ticker(response, base_market)

        return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_tickers(self, symbols: List[str], base_market: str = "KRW") -> ksxt.models.KsxtTickersResponse:
        symbol = ",".join([self.safe_symbol(base_market, symbol) for symbol in symbols])
        params = {"markets": symbol}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_tickers_price(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtTickersResponse(header=common_header, response=common_response, info=[])

        # 데이터 파싱
        parsed_info = self.parser.parse_tickers(response, base_market)

        return ksxt.models.KsxtTickersResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        params = {"markets": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_orderbook(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_orderbook(response, base_market)

        return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_orderbooks(
        self, symbols: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtMultiOrderBookResponse:
        symbol = ",".join([self.safe_symbol(base_market, symbol) for symbol in symbols])
        params = {
            #'markets': self.safe_symbol(base_market, symbol)
            "markets": symbol  # 이렇게 해야 정상동작!
        }

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_orderbooks(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtMultiOrderBookResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_orderbooks(response, base_market)

        return ksxt.models.KsxtMultiOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_balance(
        self, acc_num: str, base_market: str = "KRW", excluded_symbols: Optional[list[str]] = None
    ) -> ksxt.models.KsxtBalanceResponse:
        params = {}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_balance(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_balance(response, base_market, excluded_symbols=excluded_symbols)

        return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        params = {}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_cash(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_cash(response, base_market)

        return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        params = {"market": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = await self.public_get_fetch_security_info(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_security(response, base_market)

        return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_trade_fee(
        self, symbol: Optional[str] = "", base_market: str = "KRW"
    ) -> ksxt.models.KsxtTradeFeeResponse:
        params = {"market": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_trade_fee(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTradeFeeResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_trade_fee(response, base_market)
        return ksxt.models.KsxtTradeFeeResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_open_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtOpenOrderResponse:
        params = {
            "market": self.safe_symbol(base_market, symbol),
            "uuids": "",
            #'state': 'wait'
        }

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_opened_order(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_open_order_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_closed_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: datetime | None = None,
        end: datetime | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtClosedOrderResponse:
        params = {"market": self.safe_symbol(base_market, symbol), "uuids": "", "state": "done"}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_closed_order(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_closed_order_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def cancel_order(
        self, acc_num: str, order_id: str, symbol: str | None = "", qty: float = 0, base_market: str = "KRW", **kwargs
    ) -> ksxt.models.KsxtCancelOrderResponse:
        params = {"uuid": order_id}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_delete_send_cancel_order(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_cancel_order(response, base_market)

        return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong", "EntryShort", "ExitLong", "ExitShort"],
        otype: Literal["limit", "market"],
        price: Optional[float] = 0,
        qty: Optional[float] = 0,
        amount: Optional[float] = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:
        params = {
            "market": self.safe_symbol(base_market, symbol),
        }

        if ticket_type == "EntryLong":
            order_side = "bid"
            if otype == "limit":
                params.update({"side": order_side, "volume": qty, "price": price, "ord_type": "limit"})
                response = await self.private_post_send_order_entry(self.extend(params))

            elif otype == "market":
                params.update({"side": order_side, "price": amount, "ord_type": "price"})
                response = await self.private_post_send_order_entry_market(self.extend(params))

        elif ticket_type == "ExitLong":
            order_side = "ask"
            if otype == "limit":
                params.update({"side": order_side, "volume": qty, "price": price, "ord_type": "limit"})
                response = await self.private_post_send_order_exit(self.extend(params))

            elif otype == "market":
                params.update({"side": order_side, "volume": qty, "ord_type": "market"})
                response = await self.private_post_send_order_exit_market(self.extend(params))

        common_header = self.create_common_header(request_params=params)
        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_create_order(response, base_market)

        return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=parsed_info)

    async def fetch_withdrawal_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        params = {"state": "DONE", "uuids": "", "txids": ""}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_withdrawal_history(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtWithdrawalHistoryResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_withdrawal_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtWithdrawalHistoryResponse(
            header=common_header, response=common_response, info=parsed_info
        )

    async def fetch_deposit_history(
        self, acc_num: str, start: datetime | None = None, end: datetime | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtDepositHistoryResponse:
        params = {"state": "ACCEPTED", "uuids": "", "txids": ""}

        common_header = self.create_common_header(request_params=params)

        response = await self.private_get_fetch_deposit_history(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtDepositHistoryResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_deposit_history(
            response=response, start=start, end=end, base_market=base_market
        )

        return ksxt.models.KsxtDepositHistoryResponse(header=common_header, response=common_response, info=parsed_info)



from pydantic import BaseModel
from typing import List

from ksxt.models.common import GeneralResponse


class BalanceData(BaseModel):
    symbol: str
    name: str
    evaluation_price: float
    price: float
    pnl_amount: float
    pnl_ratio: float
    qty: float
    free_qty: float
    used_qty: float


class BalanceInfo(BaseModel):
    currency: str
    total_amount: float
    total_evaluation_amount: float
    total_pnl_amount: float
    total_pnl_ratio: float
    balances: List[BalanceData]


# Balance에 대한 구체적인 Response 타입
class KsxtBalanceResponse(GeneralResponse[BalanceInfo]):
    pass



from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel

from ksxt.models.common import GeneralResponse


class TransactionInfo(BaseModel):
    # 주문 고유 아이디
    uuid: str
    # 주문 종류 (ask, bid, deposit, withdrawal)
    type: str

    # 금액
    amount: float
    # 세금
    tax: Optional[float] = 0
    # 수수료
    fee: Optional[float] = 0

    # 계좌 번호
    account_id: Optional[str] = None
    # 화폐 통화 정보
    currency: Optional[str] = None

    # 주문 생성 시간
    created_at: datetime


class OpenedOrderInfo(TransactionInfo):
    # 종목 정보
    symbol: str
    # 가격
    price: float
    # 수량
    qty: float
    # 주문 방식
    order_type: Optional[str]


class ClosedOrderInfo(TransactionInfo):
    # 종목 정보
    symbol: str
    # 가격
    price: float
    # 수량
    qty: float
    # 주문 방식
    order_type: Optional[str]


class WithdrawalHistory(BaseModel):
    history: List[TransactionInfo]


class DepositHistory(BaseModel):
    history: List[TransactionInfo]


class OpenedOrderHistory(BaseModel):
    history: List[OpenedOrderInfo]


class ClosedOrderHistory(BaseModel):
    history: List[ClosedOrderInfo]


class KsxtWithdrawalHistoryResponse(GeneralResponse[WithdrawalHistory]):
    pass


class KsxtDepositHistoryResponse(GeneralResponse[DepositHistory]):
    pass


class KsxtOpenOrderResponse(GeneralResponse[OpenedOrderHistory]):
    pass


class KsxtClosedOrderResponse(GeneralResponse[ClosedOrderHistory]):
    pass



import logging


default_logger = logging.getLogger("")
default_logger.level = logging.INFO


class KsxtLogger:
    logger: logging.Logger = default_logger

    def _logger_ready(self, logger: logging.Logger):
        pass

    def _emit_logger(self, logger: logging.Logger = None):
        if logger:
            n = logger != getattr(self, "logger", None)
            self.logger = logger
        else:
            self.logger = default_logger
            n = True

        for obj in self.__dict__.values():
            if isinstance(obj, KsxtLogger):
                obj._emit_logger(self.logger)

        if n:
            self._logger_ready(logger)  # type: ignore



from typing import Any, Generic, Optional, TypeVar
from pydantic import BaseModel, Field
from datetime import datetime, UTC


class CommonResponseHeader(BaseModel):
    # 요청 시간
    request_time: datetime = datetime.now(UTC)
    # 요청 Parameter
    request_params: dict


class CommonResponse(BaseModel):
    # response code (success / fail)
    success: Optional[str] = None
    # response message code
    msg_code: Optional[str] = None
    # response message
    msg: Optional[str] = None
    # 거래소 별 API 호출 시 반환되는 원본 Response 데이터
    info: Any


# 제네릭을 사용한 전체 Response 모델
T = TypeVar("T")


class GeneralResponse(BaseModel, Generic[T]):
    header: CommonResponseHeader
    response: CommonResponse
    info: Optional[T] = None



import types


class Entry:
    def __init__(self, security_type, path, config):
        # equity, derivative, oversea_equity, oversea_derivative
        self.security_type = security_type

        # function key
        self.path = path

        self.config = config

        def unbound_method(_self, params={}):
            return _self.request(self.path, self.security_type, params, config=self.config)

        self.unbound_method = unbound_method

    def __get__(self, instance, owner):
        if instance is None:
            return self.unbound_method
        else:
            return types.MethodType(self.unbound_method, instance)

    def __set_name__(self, owner, name):
        self.name = name



from ksxt.api import Entry


class ImplicitAPI:

    # stock API methods

    # public methods

    # get requests
    public_get_fetch_markets = Entry("stock", "fetch_markets", {})
    public_get_fetch_security_ohlcv_minute = Entry("stock", "fetch_security_ohlcv_minute", {})
    public_get_fetch_security_ohlcv_day = Entry("stock", "fetch_security_ohlcv_day", {})
    public_get_fetch_security_ohlcv_week = Entry("stock", "fetch_security_ohlcv_week", {})
    public_get_fetch_security_ohlcv_month = Entry("stock", "fetch_security_ohlcv_month", {})
    public_get_fetch_ticker_price = Entry("stock", "fetch_ticker_price", {})
    public_get_fetch_tickers_price = Entry("stock", "fetch_tickers_price", {})
    public_get_fetch_orderbook = Entry("stock", "fetch_orderbook", {})
    public_get_fetch_orderbooks = Entry("stock", "fetch_orderbooks", {})
    public_get_fetch_security_info = Entry("stock", "fetch_security_info", {})

    # private methods

    # get requests
    private_get_fetch_balance = Entry("stock", "fetch_balance", {})
    private_get_fetch_cash = Entry("stock", "fetch_cash", {})
    private_get_fetch_trade_fee = Entry("stock", "fetch_trade_fee", {})
    private_get_fetch_opened_order = Entry("stock", "fetch_opened_order", {})
    private_get_fetch_opened_order_detail = Entry("stock", "fetch_opened_order_detail", {})
    private_get_fetch_closed_order = Entry("stock", "fetch_closed_order", {})
    private_get_fetch_closed_order_detail = Entry("stock", "fetch_closed_order_detail", {})
    private_get_fetch_withdrawal_history = Entry("stock", "fetch_withdrawal_history", {})
    private_get_fetch_deposit_history = Entry("stock", "fetch_deposit_history", {})

    # post requests
    private_post_send_order_entry = Entry("stock", "send_order_entry", {})
    private_post_send_order_entry_market = Entry("stock", "send_order_entry_market", {})
    private_post_send_order_exit = Entry("stock", "send_order_exit", {})
    private_post_send_order_exit_market = Entry("stock", "send_order_exit_market", {})

    # delete requests
    private_delete_send_cancel_order = Entry("stock", "send_cancel_order", {})



from pydantic import BaseModel

from ksxt.models.common import GeneralResponse


class CashInfo(BaseModel):
    currency: str
    cash: float
    cash_d1: float
    cash_d2: float


# Cash에 대한 구체적인 Response 타입
class KsxtCashResponse(GeneralResponse[CashInfo]):
    pass
